/***************************************************************************
 *   Copyright (C) 2004 by Sylvain Joyeux                                  *
 *   sylvain.joyeux@m4x.org                                                *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "dllistitem.h"
#include "utils.h"

#include <kglobal.h>
#include <kconfig.h>
#include <qpainter.h>

#include <libkdeapt/progress.h>

#include <klocale.h>
#include <kdebug.h>


DLListItem::DLListItem(KListView* view,
  const QString& description, const QString& download_info, int size, bool showall)
  : KListViewItem(view), m_needsRepaint(true),
    m_state(-1), m_showall(showall),
    m_description(description), m_dlinfo(download_info),
    m_current(0), m_size(size), m_progress(false)

{
  setVisible(false);
  if (s_config.empty())
    loadConfig();

  computeHeight();

  m_timestamp = QTime::currentTime();
}

DLListItem::~DLListItem() {}

void DLListItem::updateContents()
{
	m_needsRepaint = true;
	repaint();
}

void DLListItem::computeHeight()
{
  QListView* view = listView();
  const int m = view -> itemMargin();
  QFontMetrics fm(view -> viewport()->fontMetrics());
  int height = 1 + m_description.contains('\n') + m_text.size();
  if (m_progress)
    height += 1 + (m_dlinfo.isEmpty() ? 0 : 1);

  setHeight((fm.height() + 2*m) * height);
}

void DLListItem::setup()
{
  QListViewItem::setup();
  computeHeight();
}

void DLListItem::setProgression(int current, int size)
{
  m_current = current;
  m_size = size;

  updateContents();
}

int DLListItem::getState() const { return m_state; }
void DLListItem::setState(int state)
{
  if (m_state == state) return;

  m_state = state;
  if (!s_config.contains(m_state))
  {
    kdDebug() << "[setState] Unknown state " << state << ". Falling back to default mode" << endl;
    state = -1;
  }

  const Config& config(s_config[m_state]);
  setVisible(config.show_always || m_showall);
  if (config.has_progress)
  {
    setProgression(0, m_size);
    m_progress = true;
    computeHeight();
  } else if (m_progress) {
    m_progress = false;
    computeHeight();
  }

	widthChanged();
  updateContents();
}

void DLListItem::setDescription(const QString& desc)
{
  m_description = desc;
	widthChanged();
	updateContents();
}

void DLListItem::setDLInfo(const QString& dlinfo)
{
  bool new_height = m_dlinfo.isEmpty() ^ dlinfo.isEmpty();
  m_dlinfo = dlinfo;
  if (new_height) computeHeight();
	widthChanged();
	updateContents();
}

void DLListItem::setSize(unsigned long size)
{
  m_size = size;
	widthChanged();
	updateContents();
}

void DLListItem::addText(const QString& text)
{
  kdDebug() << "[DownloadItem] adding " << text << endl;
  QStringList lines(QStringList::split('\n', text));
  m_text += lines;

  computeHeight();
	widthChanged();
  updateContents();
}

void DLListItem::setShowAll(bool show)
{
  if (s_config.contains(m_state))
    setVisible(show || s_config[m_state].show_always);
  else
    setVisible(true);
  m_showall = show;
}


/***********************************************************************************
 *
 *   Drawing routines configuration
 *
 */

QMap<int, DLListItem::Config> DLListItem::s_config;

/** Adds a new config object to s_config, using the KConfig object to load
 * the color, allowing us to make colors configuratble  */
void DLListItem::loadConfigItem(KConfig* c, int state, const QString& key,
  const QColor& def_color, int order, bool show_always, bool has_progress)
{
  s_config[state] =
    Config( i18n(key), c->readColorEntry(key.lower(), &def_color), order, show_always, has_progress);
}

void DLListItem::loadConfig()
{
  using namespace APT;

  KConfig* c = KGlobal::config();

  s_config.clear();

  // This is the default config. It is a fallback when a state has no configuration.
  loadConfigItem(c, -1, I18N_NOOP("Internal error : no text defined"), QColor("gray"), 0, true, false );

  loadConfigItem(c, SetItemState::Error, I18N_NOOP("Error"), QColor("#FF0000"), 0, true, false );

  loadConfigItem(c, SetItemState::Pending, I18N_NOOP("Pending"), QColor("#954ED7"), 6, false, false );
  loadConfigItem(c, SetItemState::Get, I18N_NOOP("Get"), QColor("#009DFF"), 6, true, true );
  loadConfigItem(c, SetItemState::Hit, I18N_NOOP("Ready"), QColor("#1AE100"), 8, false, false );
  loadConfigItem(c, SetItemState::Ignore, I18N_NOOP("Ignored"), QColor("#FFDD00"), 7, false, false );
  loadConfigItem(c, SetItemState::Downloaded, I18N_NOOP("Ready"), QColor("#1AE100"), 8, false, false );

  loadConfigItem(c, SetItemState::Unpack, I18N_NOOP("Unpacking"), QColor("#009DFF"), 1, true, false );
  loadConfigItem(c, SetItemState::Setup, I18N_NOOP("Setup"), QColor("#009DFF"), 1, true, false );
  loadConfigItem(c, SetItemState::Installed, I18N_NOOP("Installed"), QColor("#1AE100"), 10, false, false );

  loadConfigItem(c, SetItemState::Remove, I18N_NOOP("Removing"), QColor("#009DFF"), 1, true, false );
  loadConfigItem(c, SetItemState::Purge, I18N_NOOP("Purging"), QColor("#009DFF"), 1, true, false );
  loadConfigItem(c, SetItemState::Removed, I18N_NOOP("Removed"), QColor("#FFDD00"), 10, false, false );

}

int DLListItem::getTabLength(const QFontMetrics & fm) const
{ return fm.width("  "); }

int DLListItem::width( const QFontMetrics & fm, const QListView * list, int /*c*/ ) const
{
	int w = 0;

	if (m_progress && !m_dlinfo.isEmpty())
		w = fm.width(m_dlinfo);

	// Compute the description size
  QStringList d_lines = QStringList::split("\n", m_description, true);
  if (m_size) d_lines.first() += " [" + sizeToString(m_size) + "]";

	for (QStringList::ConstIterator i = d_lines.begin(); i != d_lines.end(); ++i)
	{
		int new_w = fm.width(*i);
		if (w < new_w) w = new_w;
	}

	// then add the state text width
  const Config& config(s_config[m_state]);
	w += fm.width(config.text);

	// Check the text
	for (QStringList::ConstIterator i = m_text.begin(); i != m_text.end(); ++i)
  {
		int new_w = fm.width(*i);
		if (w < new_w) w = new_w;
  }

	// finally, add the margins
	return w + 2 * getTabLength(fm) + list->itemMargin();
}

/** Draws the cell
 * the cell layout is as follow:
 *        [state text] [m_description         ]
 *                     [m_dlinfo if m_progress ]
 *          [progress bar if m_progress       ]
 *             [m_text if not empty           ]
 *             [                              ]
 *
 * The state text is drawn using the color specified in the current state
 * config structure (in s_config).
 *
 * The progress bar is filled with the highlight color and outlined with
 * the text color
 *
 * Note that m_description can be multi-lines
 */

void DLListItem::paintCell(QPainter* painter, const QColorGroup& cg, int /*column*/, int w, int /*alignment*/)
{
  const Config& config(s_config[m_state]);
  const int m = listView() -> itemMargin();
	const int h = height();
  QBrush backBrush = listView() -> viewport() -> backgroundBrush();

	if (m_pixmap.isNull())
		m_pixmap.resize(w, h);

	int old_w = m_pixmap.width(), old_h = m_pixmap.height();
	if (!m_needsRepaint && (old_w != w || old_h != h))
	{
		m_pixmap.resize(w, h);
		QPainter p(&m_pixmap, this);

		if (old_w < w)
			p.fillRect(old_w, 0, w - old_w, h, backBrush);
		if (old_h < h)
			p.fillRect(0, old_h, w, h - old_h, backBrush);
		p.end();
	}
	else if (m_needsRepaint || w != m_pixmap.width() || h != m_pixmap.height())
	{
		m_pixmap.resize(w, h);
		QPainter p(&m_pixmap, this);

		p.fillRect(0, 0, w, h, backBrush);

		QFontMetrics fm(p.fontMetrics());

		// the height of a text line, margins included
		const int dy = 2 * m + fm.height();
		// the width of a "tab"
		const int tablength = getTabLength(fm);

		int y = 0;

		QRect br; // the bounding rectangle, needed to compute
							// the margin for m_description and m_moreinfo

		// Paints the state text
		p.setPen(config.color);
		p.drawText(m, y + m, w, fm.height(), AlignLeft, config.text, -1, &br);

		// Paints description
		p.setPen(cg.text());

		QStringList d_lines = QStringList::split("\n", m_description, true);
		if (m_size) d_lines.first() += " [" + sizeToString(m_size) + "]";

		for (QStringList::ConstIterator l = d_lines.begin(); l != d_lines.end(); ++l)
		{
			p.drawText(br.width() + tablength, y + m, w, fm.height(), AlignLeft, *l);
			y += dy;
		}

		if (m_progress)
		{
			if (!m_dlinfo.isEmpty())
			{
				p.drawText(br.width() + tablength, y + m, w, fm.height(), AlignLeft, m_dlinfo);
				y += dy;
			}

			QRect inside(m, y + m, w - 2*m, fm.height());

			inside.addCoords(2, 2, -2, -2);
			p.drawRect(inside);

			if (m_size)
			{
				inside.addCoords(2, 2, -2, -2);
				inside.setWidth( (int) (((float)inside.width()) * m_current / m_size) );
				p.fillRect(inside, cg.highlight());
			}
			y += dy;
		}

		for (QStringList::ConstIterator i = m_text.begin(); i != m_text.end(); ++i)
		{
			p.drawText(m + tablength, y + m, w, fm.height(), AlignLeft, *i);
			y += dy;
		}

		p.end();

		m_needsRepaint = false;
	}

	// Finally, draw onto the real painter
	painter -> drawPixmap(0, 0, m_pixmap);

}


/*********************************************************************************
 * Sorting routines
 */
template<typename T>
static int compare(T a, T b)
{
  if (a < b) return -1;
  if (a > b) return 1;
  return 0;
}
int DLListItem::compare(QListViewItem* i, int col, bool ascending) const
{
  DLListItem* item = dynamic_cast<DLListItem*>(i);
  if (!item) return KListViewItem::compare(i, col, ascending);

  if (!s_config.contains(m_state) || !s_config.contains(item->m_state))
    return KListViewItem::compare(i, col, ascending);

  int
    here = s_config[m_state].order,
    there = s_config[item -> m_state].order;

  int c = ::compare(here, there);
  if (c) return c;

  // we want items with messages first
  c = ::compare(m_text.size() ? 0 : 1, item->m_text.size() ? 0 : 1);
  if (c) return c;

  // put old items first, to avoid list flickering
  c = ::compare(m_timestamp, item->m_timestamp);
  if (c) return c;

  return QString::compare(m_description, item -> m_description);
}


