/***************************************************************************
 *   Copyright (C) 2004 by Sylvain Joyeux                                  *
 *   sylvain.joyeux@m4x.org                                                *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "getwidget.h"

#include "dllistitem.h"
//#include "boldlistitem.h"
#include "textquery.h"
#include "statsdialog.h"
#include "utils.h"

#include <kglobal.h>
#include <knotifyclient.h>
#include <qdir.h>
#include <qapplication.h>
#include <krun.h>

#include <klistview.h>
#include <kmessagebox.h>
#include <klineedit.h>
#include <kpushbutton.h>
//#include <qdialog.h>
#include <qlabel.h>
#include <qwidgetstack.h>
#include <qheader.h>
//#include <kdialogbase.h>
//#include <qvbox.h>

#include <klocale.h>
#include <kdebug.h>

GetWidget::GetWidget(QWidget *parent, const char *name)
 : DCOPObject("GetClient"), GetWidgetUI(parent, name),
 m_list(0), m_showall(true), m_active(false), m_taskHasBegun(false),
 m_cancel(false), m_hasError(false), m_totalItems(0)
{
  // Build the download part
  m_list = new KListView(stack);
  stack -> addWidget(m_list);
  {
    m_list -> addColumn(i18n("Info"));
    m_list -> setAllColumnsShowFocus(true);
    m_list -> setShowSortIndicator(true);
    m_list -> setSelectionMode(QListView::NoSelection);
    m_list -> setResizeMode(KListView::NoColumn);
    m_list -> header()->hide();
  }
  stack -> raiseWidget(m_list);

  connect(m_list, SIGNAL(executed(QListViewItem*)),
    this, SLOT(itemExecuted(QListViewItem*)));
}


GetWidget::~GetWidget()
{
}

/***************************************************************
 * Status display methods
 */

void GetWidget::beginTask(const QString& description,
  const QString& end_message_success, const QString& end_message_errors, bool has_stats)
{
  m_totalItems = 0;
  m_items.clear();
  m_list -> clear();
	m_list -> adjustColumn(0);

  m_cancel = false;
  m_hasError = false;
  m_taskHasBegun = false;
	m_taskHasStats = has_stats;

  m_taskEndMsgSuccess = end_message_success;
  m_taskEndMsgError = end_message_errors;

  setActive(true);

  task -> setText(description);

	emit userInteractionBegin();
}

void GetWidget::endCleanup()
{
  emit setStatusText( "" );

  m_taskHasBegun = false;

  task -> setText(i18n("Idle"));
  stage -> setText("");
  setActive(false);
}
void GetWidget::endTask()
{
  QApplication::sendPostedEvents();

  endCleanup();

  // If the user canceled, do not do notifications
  if (m_cancel) return;

  if (m_hasError)
    KNotifyClient::event(winId(), "kpkgmanager_endtask", m_taskEndMsgError);
  else
    KNotifyClient::event(winId(), "kpkgmanager_endtask", m_taskEndMsgSuccess);
}

bool GetWidget::endStatus() const
{ return m_hasError; }

void GetWidget::displayGlobalAction(const QString& /*description*/)
{
}

void GetWidget::discardGlobalAction()
{
}

/***************************************************************
 * User interaction
 */
void GetWidget::update()
{
  if (isActive()) return;

  beginTask(i18n("Updating the list of available packages"),
    i18n("The package lists update finished successfully"),
    i18n("The package lists update finished with errors"), false);

  if (!GetClient::update())
    endCleanup();
}

void GetWidget::upgrade()
{
  if (isActive()) return;

  beginTask(i18n("System upgrade"),
    i18n("The system upgrade finished successfully"),
    i18n("The system upgrade finished with errors"));

  if (!GetClient::upgrade(false))
    endCleanup();
}

void GetWidget::distupgrade()
{
  if (isActive()) return;
  beginTask(i18n("System dist-upgrade"),
    i18n("The system dist-upgrade finished successfully"),
    i18n("The system dist-upgrade finished with errors"));

  if (!GetClient::upgrade(true))
    endCleanup();
}

void GetWidget::install(const QString& commands)
{
  if (isActive()) return;

  beginTask(i18n("Package installation and removal"),
    i18n("The install & remove actions have finished successfully"),
    i18n("The install & remove actions have finished with errors"));
  QStringList query = QStringList::split(' ', commands);
  for (QStringList::Iterator i = query.begin(); i != query.end(); ++i)
  {
    bool has_command = (*i).startsWith("+") || (*i).startsWith("-");
    if (! has_command)
      (*i) = "+" + (*i);
  }

  if (!GetClient::install(query))
    endCleanup();
}


void GetWidget::setShowAll(bool show)
{
  if (show == m_showall) return;
  m_showall = show;

  for (Items::ConstIterator i = m_items.begin(); i != m_items.end(); ++i)
    (*i) -> setShowAll(m_showall);
}

/**************************************************************************************
 * APT::Get interface
 */

void GetWidget::beginTask()
{
  m_taskHasBegun = true;

	// Autohide if we're not waiting for the stats dialog
	if (!m_taskHasStats)
		emit userInteractionEnd();
}

void GetWidget::beginStage(int category, QString stageDescription)
{
  stage -> setText(stageDescription);
  emit setStatusText( "" );
  // We are called from GetClient _and_ the Get DCOP object
  if (needsAnswer()) setQueryAnswer(true);
}

void GetWidget::newItem(QString key, QString desc, QString dlinfo, int size)
{
  kdDebug() << "New item : " << key << " " << desc << " " << dlinfo << " " << size << endl;

  if (m_items.contains(key))
  {
    // Reuse the item, but change the description
    DLListItem* item = m_items[key];
    item->setDescription(desc);
    item->setDLInfo(dlinfo);
    if (size) item->setSize(size);
    return;
  }

  kdDebug() << "adding ..." << endl;
  m_pending[key].desc = desc;
  m_pending[key].dlinfo = dlinfo;
  m_pending[key].size = size;
}

void GetWidget::setItemState(QString key, int state)
{
  m_hasError = m_hasError || (state == APT::SetItemState::Error);

  kdDebug() << "SetItemState : " << key << " " << state << endl;
  if (m_pending.contains(key))
  {
    const DLPendingItem& item(m_pending[key]);
    m_items[key] = new DLListItem(m_list, item.desc, item.dlinfo, item.size, m_showall);
    for (QStringList::ConstIterator it = item.text.begin(); it != item.text.end(); ++it)
      m_items[key] -> addText(*it);
    m_pending.remove(key);
  }

  if (m_items.contains(key))
    m_items[key] -> setState(state);
}

void GetWidget::addItemText(const QString& key, const QString& text)
{
  if (m_pending.contains(key))
    m_pending[key].text += text;
  else if (m_items.contains(key))
    m_items[key] -> addText(text);
}

void GetWidget::itemProgression(QString key, int current, int total)
{
  if (!m_items.contains(key)) return;
  m_items[key] -> setProgression(current, total);
}

void GetWidget::questionYN(QString question)
{
  emit userInteractionBegin();
  setQueryAnswer(KMessageBox::questionYesNo(this, question, i18n("KPkgManager")) == KMessageBox::Yes);
  emit userInteractionEnd();
}

void GetWidget::questionText( QString question )
{
  emit userInteractionBegin();

  TextQuery dialog;
  dialog.m_label -> setText(question);

  int code = dialog.exec();
  if (code == QDialog::Accepted)
  {
    setQueryAnswer(dialog.m_edit -> text());
    return;
  }

  setQueryAnswer(QString::null);

  emit userInteractionEnd();
}

void GetWidget::conffile( QString file )
{
  emit userInteractionBegin();

  static const QString message = i18n(
    "The configuration file \"%1\" has been modified by you and the "
    "package manager at the same time. Do you want to overwrite "
    "your version with the one in the package ?");

  int code = KMessageBox::warningYesNo(this, message.arg(file), "APT/Configuration file change");
  if (code == KMessageBox::Yes)
  {
    const QString file (file);

    QDir dir;
    static const QString old_suffix(".dpkg-old"), new_suffix(".dpkg-new");

    if (!dir.exists(file))
      kdDebug() << "Inexistent configuration file " << file << endl;
    else if (!dir.exists(file + new_suffix))
      kdDebug() << "Inexistent configuration file " << file + new_suffix << endl;
    else
    {
      dir.rename(file, file + old_suffix);
      dir.rename(file + new_suffix, file);
    }
  }

  setQueryAnswer(QVariant());

  emit userInteractionEnd();
}


/****************************************************************************************
 * Receiving install/remove/upgrade stats from the APT backend
 */

void GetWidget::stats( APT::Stats event )
{
	// It has already been sent by GetWidget::beginTask
  // emit userInteractionBegin();

  StatsDialog* dialog = new StatsDialog(this);

  bool accepted = dialog->exec(event);
  if (accepted)
  {
    const StatsDialog::ItemList& items(dialog->getItems());
    for (QValueList<StatsDialog::Item>::ConstIterator i = items.begin(); i != items.end(); ++i)
    {
      newItem((*i).package, (*i).package, "", (*i).size);
      setItemState((*i).package, (*i).state);
    }
  }
  else
  {
    m_cancel = true;
  }

  setQueryAnswer(accepted);
  delete dialog;
  emit userInteractionEnd();
}

/************************************************************************************
 * Notification messages
 */

void GetWidget::error( QString key, QString message )
{
  kdDebug() << "error " << key << " " << message << endl;

  m_hasError = true;
  if (key.isEmpty())
  {
    emit userInteractionBegin();
    KMessageBox::error(this, message, i18n("KPkgManager error"));
  }
  else
  {
    addItemText(key, message);
    setItemState(key, APT::SetItemState::Error);
  }
}

void GetWidget::warning( QString key, QString message )
{
  if (key.isEmpty())
  {
    emit userInteractionBegin();
    KMessageBox::error(this, message, i18n("KPkgManager warning"));
    emit userInteractionEnd();
  }
  else
    addItemText(key, message);
}

void GetWidget::info( QString key, QString message )
{
  if (key.isEmpty())
  {
    emit userInteractionBegin();
    KMessageBox::information(this, message, i18n("Information"));
    emit userInteractionEnd();
  }
  else
    addItemText(key, message);
}

void GetWidget::overallProgression(int curItem, int totalItems,
    double /*curBytes*/, double /*totalBytes*/, double cps)
{
  if (curItem == totalItems && m_totalItems)
    totalItems = m_totalItems;

  QStringList message;
  if (curItem != totalItems)
    message += i18n("%1/%2 items ready").arg(curItem).arg(totalItems);
  else if (curItem)
    message += i18n("%1 items ready").arg(curItem);

  if (cps)
    message += QString("(%1/s)").arg(sizeToString(cps));

  emit setStatusText(message.join(" "));
}

void GetWidget::subwin(int id)
{
/*  QXEmbed* embed = new QXEmbed(this);
  embed -> setProtocol(QXEmbed::XPLAIN);
  embed -> embed(id);*/
//  bla
}

void GetWidget::exit( )
{
  if (!isActive()) return;

  if (m_taskHasBegun)
  {
    kdDebug() << "The package manager died !!!" << endl;
    error(QString::null, i18n("The package manager has exited unexpectedly"));
  }

  endCleanup();
}

void GetWidget::setActive(bool is_active)
{
  m_active = is_active;
  emit active(is_active);
}
bool GetWidget::isActive( ) const
{ return m_active; }

void GetWidget::cancel( )
{
  m_cancel = true;
  cancelCurTask(false);
}

void GetWidget::itemExecuted(QListViewItem* item)
{
/*  if (!item) return;

  Items::Iterator dlitem;
  for (dlitem = m_items.begin(); dlitem != m_items.end(); ++dlitem)
    if ((*dlitem) == item) break;
  if (dlitem == m_items.end()) return;

  if (dlitem.key().at(0).isDigit())
    return;

  new KRun( QString("apt:/show&").append(dlitem.key()) );*/
}

#include "getwidget.moc"
