/*
 * Copyright (C) 2004 Sylvain Joyeux <sylvain.joyeux@m4x.org>
 */

#include "kpkgmanager.h"

#include "getwidget.h"
#include "pkgmanagerdcop.h"

#include "settings.h"
#include "settingsbehaviorui.h"
#include <kconfigdialog.h>
#include <kconfig.h>

#include <kstatusbar.h>
#include <kaction.h>
#include <knotifyclient.h>
#include <knotifydialog.h>
#include <kapplication.h>
// #include <kstdaction.h>
#include <dcopclient.h>
#include <qapplication.h>
#include <kaction.h>

#include <kpushbutton.h>
#include "textquery.h"
#include <klineedit.h>
#include <ksystemtray.h>
#include <kpopupmenu.h>
#include <qtooltip.h>
#include <kdialogbase.h>
#include <kwin.h>

#include <klocale.h>
#include <kdebug.h>

using namespace APT;
using namespace KNotifyClient;

KPkgManager::KPkgManager()
  : KMainWindow( 0 ), m_systray(0),
  m_autohide(0), m_userInteractionSection(false)
{
  // set the shell's ui resource file
  setXMLFile("kpkgmanagerui.rc");

  m_central = new GetWidget(this);
  setCentralWidget(m_central);

  connect(m_central, SIGNAL(setStatusText(const QString&)),
    this, SLOT(setGetWidgetStatus(const QString&)));
  connect(m_central, SIGNAL(userInteractionBegin(const QString&)),
    this, SLOT(userInteractionBegin(const QString&)));
  connect(m_central, SIGNAL(userInteractionEnd(const QString&)),
    this, SLOT(userInteractionEnd(const QString&)));
  connect(m_central, SIGNAL(active(bool)), this, SLOT(isActive(bool)));

  setupActions();
  createGUI();

  setAutoSaveSettings();

  statusBar() -> insertItem("", 0, 1); // Mode string

  m_dcop = new PkgManagerDCOP(this);

  if (Settings::systray())
    setupSystray();

  if (Settings::startHidden() && Settings::systray())
    hide();
  else show();

  buildSystrayTip("");
}

KPkgManager::~KPkgManager()
{
  delete m_dcop;
  delete m_systray;
}

void KPkgManager::setupActions()
{
  m_actUpdate=new KAction(i18n("Update"), 0, this, SLOT(update()),
      actionCollection(), "get_update");
  m_actUpgrade=new KAction(i18n("Upgrade"), 0, this, SLOT(upgrade()),
      actionCollection(), "get_upgrade");
  m_actDistUpgrade=new KAction(i18n("Dist-upgrade"), 0, this, SLOT(distupgrade()),
      actionCollection(), "get_distupgrade");
  m_actInstall=new KAction(i18n("Install/Remove"), 0, this, SLOT(install()),
      actionCollection(), "get_install");

  KStdAction::configureNotifications(this, SLOT(configureNotifications()), actionCollection());
  KStdAction::preferences(this, SLOT(configure()), actionCollection());

  KToggleAction* toggle;

  toggle = KStdAction::showStatusbar(0, 0, actionCollection(), 0);
  connect(toggle, SIGNAL(toggled(bool)), statusBar(), SLOT(setShown(bool)));

  toggle = new KToggleAction(i18n("Show all items"), 0, actionCollection(), "options_show_all");
  connect(toggle, SIGNAL(toggled(bool)), this, SLOT(setShowAll(bool)));
  bool showall = Settings::showAll();
  toggle->setChecked(showall);
  m_central -> setShowAll(showall);

  m_actContinue =
    new KAction(i18n("Continue"), 0, 0, 0, actionCollection(), "task_continue");
  connect(m_central->taskContinue, SIGNAL(clicked()), m_actContinue, SLOT(activate()));
  connect(m_actContinue, SIGNAL(enabled(bool)), m_central -> taskContinue, SLOT(setShown(bool)));
  connect(m_actContinue, SIGNAL(enabled(bool)), m_central -> taskContinue, SLOT(setEnabled(bool)));
  m_actContinue -> setEnabled(false);

  m_actCancel =
    new KAction(i18n("Cancel"), 0, this, SLOT(cancelCurTask()), actionCollection(), "task_cancel");
  connect(m_central -> taskCancel, SIGNAL(clicked()), m_actCancel, SLOT(activate()));
  connect(m_actCancel, SIGNAL(enabled(bool)), m_central -> taskCancel, SLOT(setEnabled(bool)));
  m_actCancel -> setEnabled(false);
}

void KPkgManager::setupSystray()
{
  if (m_systray) return;

  m_systray = new KSystemTray(this);
  m_systray -> setPixmap(KSystemTray::loadIcon("kpkgmanager"));

  KPopupMenu* systray = m_systray->contextMenu();
  m_actUpdate->plug(systray);
  m_actUpgrade->plug(systray);
  m_actDistUpgrade->plug(systray);
  m_actInstall->plug(systray);

  m_actCancel -> plug(systray);

  connect(m_systray, SIGNAL(quitSelected()), KApplication::kApplication(), SLOT(quit()));
  m_systray -> show();
}

void KPkgManager::update()
{
  if (m_central->isActive()) return;
  m_central->update();
}
void KPkgManager::distupgrade()
{
  if (m_central->isActive()) return;
  m_central->distupgrade();
}
void KPkgManager::upgrade()
{
  if (m_central->isActive()) return;
  m_central->upgrade();
}
void KPkgManager::install()
{
  if (m_central->isActive()) return;

  TextQuery* edit = new TextQuery(this);
  edit -> m_label -> setText(
    i18n("List packages to install. Prefix with '-' if you"
          " want to remove a package instead of installing it"));
  if (edit -> exec() == QDialog::Rejected) return;

  m_central->install(edit -> m_edit -> text());
}

void KPkgManager::cancelCurTask()
{
  m_actCancel -> setEnabled(false);
  m_central->cancel();
}


void KPkgManager::initAutoHide()
{
  m_autohide = 0;
  if (!Settings::systray()) return;
  if (Settings::autoHideExtended())
    m_autohide = 2;
  else if (Settings::autoHide())
    m_autohide = 1;
}
void KPkgManager::resetAutoHide()
{ m_autohide = 0; }

bool KPkgManager::queryClose()
{
  if (m_systray)
  {
    hide();
    return false;
  }

  return true;
}

void KPkgManager::hideEvent(QHideEvent*)
{ initAutoHide(); }
void KPkgManager::showEvent(QShowEvent*)
{
  if (!m_userInteractionSection)
    resetAutoHide();
}

bool KPkgManager::queryExit( )
{
  // kpkgmanagerd & the kded daemon will take care
  // of cancelling running tasks
  Settings::writeConfig();
  return KMainWindow::queryExit();
}

void KPkgManager::configureNotifications()
{ KNotifyDialog::configure(this); }

void KPkgManager::buildSystrayTip(const QString& status)
{
  if (!m_systray) return;

  QString tooltip = "<b>" + m_central->task->text() + "</b>";

  QString curstage = m_central->stage->text();
  if (!curstage.isEmpty())
    tooltip += "<br>" + curstage;
    if (!status.isEmpty())
    tooltip += "<br>" + status;

  QToolTip::add(m_systray, tooltip);
}

void KPkgManager::setGetWidgetStatus(const QString& message)
{
  buildSystrayTip(message);
  statusBar() -> changeItem(message, 0);
}

GetWidget* KPkgManager::getGetWidget( ) const
{ return m_central; }

void KPkgManager::isActive(bool is_active)
{
  stateChanged("is_active", is_active ? StateNoReverse : StateReverse);
  if (!is_active)
    m_userInteractionSection = false;
  if (!is_active && m_autohide && !m_central->endStatus())
    hide();
  buildSystrayTip("");
}

void KPkgManager::setShowAll( bool show )
{
  Settings::setShowAll(show);
  m_central->setShowAll(show);
}

void KPkgManager::userInteractionBegin( const QString & message )
{
  m_userInteractionSection = true;
  show();
  KWin::forceActiveWindow(winId());
}

void KPkgManager::userInteractionEnd( const QString & message )
{
  m_userInteractionSection = false;
  if (m_autohide == 2)
    hide();
}


void KPkgManager::configure()
{
  //An instance of your dialog could be already created and could be cached,
  //in which case you want to display the cached dialog instead of creating
  //another one
  if ( KConfigDialog::showDialog( "settings" ) )
          return;

  // KConfigDialog didn't find an instance of this dialog, so lets create it :
  KConfigDialog* dialog = new KConfigDialog( this, "settings",
                                            Settings::self() );
  SettingsBehaviorUI* confWdg =
          new SettingsBehaviorUI( 0, "Settings" );

  dialog->addPage( confWdg, i18n("General behavior"), "behavior" );

  //User edited the configuration - update your local copies of the
  //configuration data
  connect( dialog, SIGNAL(settingsChanged()),
          this, SLOT(updateConfiguration()) );

  dialog->show();
}

void KPkgManager::updateConfiguration()
{
  if (m_systray && !Settings::systray())
  {
    show();
    delete m_systray;
    m_systray = 0;
    resetAutoHide();
  }
  else if (!m_systray && Settings::systray())
  {
    setupSystray();
    initAutoHide();
  }
}

void KPkgManager::saveGlobalProperties(KConfig* session)
{
  session->writeEntry("shown", isShown());
}
void KPkgManager::readGlobalProperties(KConfig* session)
{
  bool is_shown = session->readBoolEntry("shown", false);
  if (!is_shown && Settings::systray())
    hide();
  else
    show();
}

#include "kpkgmanager.moc"
