/***************************************************************************
 *   Copyright (C) 2004 by Sylvain Joyeux                                  *
 *   sylvain.joyeux@m4x.org                                                *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "statsdialog.h"

#include "statswidget.h"
#include "utils.h"
#include "textquery.h"
#include "settings.h"

#include <qlayout.h>
#include <qhbox.h>
#include <qsplitter.h>
#include <qlabel.h>
#include <klistview.h>
#include <boldlistitem.h>
#include <klineedit.h>
#include <kpushbutton.h>

#include <kactioncollection.h>
#include <ktoolbar.h>
#include <kaction.h>
#include <kstdaction.h>
#include <knotifyclient.h>

#include <khtml_part.h>
#include <khtmlview.h>
#include <libkdeapt/progress.h>

#include <kprotocolinfo.h>

#include <klocale.h>
#include <kdebug.h>

using namespace APT;

StatsDialog::StatsDialog(QWidget *parent, const char *name)
 : StatsWidget(parent, name)
{
  connect(lstPackages, SIGNAL(executed(QListViewItem *)),
    this, SLOT(packageExecuted(QListViewItem *)));

  connect(btnOk, SIGNAL(clicked()), this, SLOT(accept()));
  connect(btnCancel, SIGNAL(clicked()), this, SLOT(reject()));

  QVBoxLayout* htmlLayout = new QVBoxLayout(htmlContainer, spacingHint());

  KToolBar* toolbar = new KToolBar(htmlContainer);
  htmlLayout->addWidget(toolbar);
  toolbar->setBarPos(KToolBar::Top);
  toolbar->setMovingEnabled(false);

  m_html = new KHTMLPart(htmlContainer);
  htmlLayout->addWidget(m_html->view());
  m_html->setJScriptEnabled(false);
  m_html->setJavaEnabled(false);
  m_html->setMetaRefreshEnabled(false);
  m_html->setPluginsEnabled(false);

  connect(m_html->browserExtension(), SIGNAL(openURLRequestDelayed(const KURL&, const KParts::URLArgs&)),
    this, SLOT(openURL(const KURL&)));

  m_urlCurrent = m_urlHistory.begin();

  setupActions(toolbar);
}

void StatsDialog::setupActions(KToolBar* toolbar)
{
  m_actionCollection = new KActionCollection(this);

  QHBox* show_widget = new QHBox(toolbar);
  show_widget -> setSpacing(spacingHint());
  {
    QLabel* label = new QLabel(show_widget);
    label -> setText(i18n("Show package : "));
    m_showLineEdit = new KLineEdit(show_widget);
    m_showLineEdit->setTrapReturnKey(true);
    connect(m_showLineEdit, SIGNAL(returnPressed(const QString&)),
      this, SLOT(showPackage(const QString&)));
  }

  m_actShow = new KWidgetAction(show_widget, i18n("Show package: "),
    KShortcut(), 0, 0, m_actionCollection, "new_package");
  m_actShow->plug(toolbar);

  m_actBack = KStdAction::back(this, SLOT(urlBack()), m_actionCollection);
  m_actBack->plug(toolbar);
  m_actForward = KStdAction::forward(this, SLOT(urlForward()), m_actionCollection);
  m_actForward->plug(toolbar);

  setBrowseActionsState();
}

StatsDialog::~StatsDialog() {}

bool StatsDialog::hasKIOApt()
{ return KProtocolInfo::isKnownProtocol(KURL("apt:/")); }

void StatsDialog::setupLayout()
{
  QSize size = Settings::statsSize();
  QValueList<int> splitterSizes = Settings::statsSplitter();

  if (size.isEmpty())
  {
    size = QWidget::size();
    Settings::setStatsSize(size);
  }

  m_showHTML = hasKIOApt();
  htmlContainer->setShown(m_showHTML);
  if (!m_showHTML)
  {
    size.setWidth(splitterSizes.front());
    if (!splitterSizes.empty())
      splitterSizes.pop_back();
  }

  if (!splitterSizes.empty())
    mainSplitter->setSizes(splitterSizes);

  resize(size);
}

void StatsDialog::saveLayout()
{
  if (m_showHTML)
  {
    QSize size = this->size();
    QValueList<int> splitterSizes = mainSplitter->sizes();

    Settings::setStatsSize(size);
    Settings::setStatsSplitter(splitterSizes);
  }
}

bool StatsDialog::exec(const APT::Stats& event)
{
  fillLabels(event);
  int acting = fillTree(event);

  QString message;
  if (acting == DoNothing)
  {
    if (lstPackages->childCount() == 0)
    {
      KNotifyClient::event( static_cast<QWidget*>(parent())->winId(), "kpkgmanager_noaction");
      return true;
    }

    message = i18n("There's nothing to do");
    btnCancel->setEnabled(false);
  }
  else
    message = i18n("Do you want to continue ?");

  KNotifyClient::event(static_cast<QWidget*>(parent())->winId(), "kpkgmanager_stats",
    i18n("Ready to proceed. Please confirm"));

  lblWarning -> setText(message);

  setupLayout();

  kdDebug() << "size : " << size().width() << " x " << size().height() << endl;
  bool accepted = (KDialog::exec() == QDialog::Accepted);
  kdDebug() << "size : " << size().width() << " x " << size().height() << endl;

  saveLayout();

  if (accepted && (acting == Dangerous))
  {
    TextQuery* query = new TextQuery(this);
    QString phrase = i18n("Yes, do as I say");
    query->m_label->setText(
      i18n("You asked for something potentially harmful. Please type '%1' if you want to continue.").arg(phrase)
    );
    accepted = (query -> exec() == QDialog::Accepted);
    if (accepted)
      accepted = (query->m_edit -> text() == phrase);
  }

  return accepted;
}

enum StatModes
{
  Nontrivial = 4,
  Download = 8
};
struct StatField
{
  QString block_name;
  int packages;

  int mode;
  int priority;
};

/** Returns 0 if no action is needed, 1 if something
 * is to be done and 2 if it is potentially harmful */
int StatsDialog::fillTree(const APT::Stats& event)
{
  // list of the various fields available
  // (the format is name stats_field mode)
  static StatField fields[] = {
    { i18n("Kept back packages"), Stats::Kept, DoNothing, 1 },
    { i18n("Held packages"), Stats::Held, DoNothing, 1 },
    { i18n("Recommended packages"), Stats::Recommended, DoNothing, 1 },
    { i18n("Suggested packages"), Stats::Suggested, DoNothing, 1 },
    { i18n("Upgraded packages"), Stats::Upgraded, Normal|Download, 0 },
    { i18n("Installed packages"), Stats::Installed, Normal|Download, 0 },
    { i18n("To be configured packages"), Stats::BadState, Normal, -1 },
    { i18n("Downgraded packages"), Stats::Downgraded, Normal|Nontrivial|Download, -1 },
    { i18n("Removed packages"), Stats::Removed, Normal|Nontrivial, -1 },
    { i18n("Removed essential packages"), Stats::Essential, Normal|Nontrivial|Dangerous, -1 },
    { QString::null, 0, DoNothing, 2 }
  };

  bool acting = false, dangerous = false;
  m_itemList.clear();

  const Stats::Downloads& downloads(event.downloads);
  for (StatField* field = fields; !field -> block_name.isNull(); ++field)
  {
    const QStringList& list(event.get(field->packages));
    if (list.empty()) continue;

    bool is_nontrivial = (field->mode & Nontrivial);
    bool needs_download = (field->mode & Download);

    KListViewItem* node
        = new BoldListItem(is_nontrivial, lstPackages, field->block_name);
    node->setOpen(true);
    for (QStringList::ConstIterator i = list.begin(); i != list.end(); ++i)
    {
      new BoldListItem(event.get(Stats::Query).contains(*i), node, *i);

      if (!needs_download) continue;

      // check if the item is already downloaded or not
      Stats::DownloadIterator download_item = downloads.begin();
      for (; download_item != downloads.end(); ++download_item)
        if ((*download_item).key == *i) break;

      int state = SetItemState::Downloaded;
      unsigned long size = 0;
      if (download_item != downloads.end())
      {
        state = SetItemState::Pending;
        size = (*download_item).size;
      }

      m_itemList.append(Item(*i, state, size));
    }

    acting = acting || (field->mode & Normal);
    dangerous = dangerous || (field->mode & Dangerous);
  }

  if (!acting) return DoNothing;
  if (dangerous) return Dangerous;
  return Normal;
}

void StatsDialog::fillLabels(const APT::Stats& event)
{
  QString message;

  if (event.unpacked >= 0)
    message = i18n("After unpacking, %1 of disk space will be used")
        .arg(sizeToString(event.unpacked));
  else
    message = i18n("After unpacking, %1 of disk space will be freed")
        .arg(sizeToString(-event.unpacked));
  lblUnpacked -> setText(message);

  QString fetch, total;
  fetch = sizeToString(event.fetch);
  total = sizeToString(event.total);
  if (event.fetch != event.total)
    message = i18n("Need to get %1/%2 of archives").arg(fetch).arg(total);
  else
    message = i18n("Need to get %1 of archives").arg(fetch);

  lblGet -> setText(message);
}

const StatsDialog::ItemList& StatsDialog::getItems( ) const
{ return m_itemList; }

void StatsDialog::packageExecuted(QListViewItem* item)
{
  QString text = item->text(0);
  bool hasParent = item->parent();

  // Non top levels items are not packages
  if (!hasParent) return;

  m_showLineEdit->setText("");

  // thanks to debian, there's no spaces in package names
  QString package = text.section(" ", 0, 0);
  showPackage(package);
}

void StatsDialog::showPackage(const QString& package)
{
  KURL url("apt:/show");
  url.setQuery(package);
  url.addQueryItem("enable_actions", "0");
  url.addQueryItem("enable_search", "0");
  url.addQueryItem("stay_internal", "1");
  openURL(url);
}

void StatsDialog::openURL(const KURL& url)
{
  if (!m_urlHistory.empty())
  {
    while(m_urlCurrent != m_urlHistory.fromLast())
      m_urlHistory.pop_back();
  }

  m_urlCurrent = m_urlHistory.append(url);
  m_html->openURL(url);

  setBrowseActionsState();
}

void StatsDialog::urlBack()
{
  if (m_urlHistory.empty()) return;
  if (m_urlCurrent == m_urlHistory.begin())
    return;

  --m_urlCurrent;
  m_html->openURL(*m_urlCurrent);

  setBrowseActionsState();
}

void StatsDialog::urlForward()
{
  if (m_urlHistory.empty()) return;
  if (m_urlCurrent == m_urlHistory.fromLast())
    return;

  ++m_urlCurrent;
  m_html->openURL(*m_urlCurrent);

  setBrowseActionsState();
}

void StatsDialog::setBrowseActionsState()
{
  bool forward = false, back = false;
  if (!m_urlHistory.empty())
  {
    forward = (m_urlCurrent != m_urlHistory.fromLast());
    back = (m_urlCurrent != m_urlHistory.begin());
  }

  m_actBack->setEnabled(back);
  m_actForward->setEnabled(forward);
}

#include "statsdialog.moc"
