/***************************************************************************
 *   Copyright (C) 2004 by Sylvain Joyeux                                  *
 *   sylvain.joyeux@m4x.org                                                *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "getclient.h"

#include <kapplication.h>

#include <klocale.h>
#include <kdebug.h>

/** kpkgmanagerd::Get signals */
const QCString GetClient::s_getSignals[] =
{
  "beginTask()",
// endTask is called by pmEndTask
//  "endTask()",
  "beginStage(int,QString)",
  "newItem(QString,QString,QString,int)",
  "setItemState(QString,int)",
  "itemProgression(QString,int,int)",
  "questionYN(QString)",
  "questionText(QString)",
  "conffile(QString)",
  "stats(APT::Stats)",
  "overallProgression(int,int,double,double,double)",
  "error(QString,QString)",
  "warning(QString,QString)",
  "info(QString,QString)",
  "subwin(int)",
  0
};


GetClient::GetClient()
  : m_kdedRef("kded", "pkgmanager"),
    m_connected(false), m_pendingArg(false)
{
  // KDED module signals
  connectDCOPSignal(m_kdedRef.app(), m_kdedRef.obj(), "pmIsReady()", "pmIsReady()", false);
}

GetClient::~GetClient() {}

bool GetClient::lock()
{
  beginStage(APT::BeginStage::PreRun, i18n("Waiting the package manager daemon"));

  DCOPReply call_reply = m_kdedRef.call("lock");
  if (!call_reply.isValid())
  {
    error(QString::null, i18n("Unable to talk to the package manager KDED daemon. Please check your installation."));
    return false;
  }

  if (! (bool)call_reply)
  {
    m_kdedRef.call("unlock");
    error(QString::null, readError());
    return false;
  }

  call_reply = m_kdedRef.call("pmGet");
  call_reply.get(m_getRef);
  connectDCOPSignal(m_kdedRef.app(), m_kdedRef.obj(), "pmExit()", "pmExit()", true);
//
  return true;
}

void GetClient::connectSignals()
{
  if (m_connected) return;

  kdDebug() << "[GetClient] Connecting signals" << endl;
  for (const QCString* sig = s_getSignals; !sig->isNull(); ++sig)
  {
    if (!connectDCOPSignal(m_getRef.app(), m_getRef.obj(), *sig, *sig, true))
    {
      kdDebug() << "Unable to connect " << m_getRef.app() << "/" << m_getRef.obj() << "."
          << *sig << endl;
    }
  }

  connectDCOPSignal(m_getRef.app(), m_getRef.obj(), "beginTask()", "pmBeginTask()", true);
  connectDCOPSignal(m_getRef.app(), m_getRef.obj(), "endTask()", "pmEndTask()", true);

  m_connected = true;
}

void GetClient::disconnectSignals()
{
  if (!m_connected) return;

  for (const QCString* sig = s_getSignals; !sig->isNull(); ++sig)
    disconnectDCOPSignal(m_getRef.app(), m_getRef.obj(), *sig, *sig);

  disconnectDCOPSignal(m_getRef.app(), m_getRef.obj(), "beginTask()", "pmBeginTask()");
  disconnectDCOPSignal(m_getRef.app(), m_getRef.obj(), "endTask()", "pmEndTask()");

  m_connected = false;
}

void GetClient::pmIsReady()
{
  if (m_pending.isEmpty()) return;

  kdDebug() << "Calling kpkgmanagerd" << m_pending << endl;

  connectSignals();

  bool ret;
  QCString f = m_pending;
  m_pending = QString::null;
  if (m_pendingHasArg)
    ret = (bool) m_getRef.call(f, m_pendingArg);
  else
    ret = (bool) m_getRef.call(f);

  if (!ret)
  {
    disconnectSignals();
    error(QString::null, readError());
  }
}

void GetClient::setupCall(const QCString& function)
{
  m_pending = function;
  m_pendingHasArg = false;
}
template<class T>
void GetClient::setupCall(const QCString& function, const T& arg)
{
  m_pending = function;
  m_pendingHasArg = true;
  m_pendingArg = DCOPArg(arg);
}

bool GetClient::install(const QStringList& query)
{
  setupCall("install", query);
  return lock();
}

bool GetClient::update()
{
  setupCall("update");
  return lock();
}

bool GetClient::upgrade(bool distupgrade)
{
  setupCall("upgrade", distupgrade);
  return lock();
}

QString GetClient::readError()
{
  DCOPReply r = m_kdedRef.call("error");
  QString error; r.get(error);
  if (error.isEmpty())
  {
    return i18n("It seems that there is an error with no message"
                " explaining it. Please report this as a bug");
  }
  return error;
}

void GetClient::pmBeginTask() {};
void GetClient::pmEndTask()
{
  disconnectSignals();
  m_kdedRef.call("unlock");
  disconnectDCOPSignal(m_kdedRef.app(), m_kdedRef.obj(), "pmExit()", "pmExit()");

  endTask();
}

void GetClient::pmExit()
{
  m_pending = QString::null;
  disconnectSignals();
  exit();
}

void GetClient::setQueryAnswer( QVariant answer )
{
  m_getRef.call("setQueryAnswer", answer);
}

bool GetClient::needsAnswer( )
{
  if (m_getRef.isNull()) return false;
  DCOPReply ret = m_getRef.call("needsAnswer");
  if (!ret.isValid()) return false;
  return (bool) m_getRef.call("needsAnswer");
}

void GetClient::cancelCurTask(bool wait)
{
  if (m_pending.isEmpty())
    m_getRef.call("cancelCurTask", wait);
  else
    m_kdedRef.call("cancel", wait);
}

