/***************************************************************************
 *   Copyright (C) 2004 by Sylvain Joyeux                                  *
 *   sylvain.joyeux@m4x.org                                                *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef GETCLIENT_H
#define GETCLIENT_H

#include <dcopobject.h>
#include <dcopref.h>
#include <qstringlist.h>
#include <qtimer.h>

#include <libkdeapt/progress.h>

/** Base class for clients to the kpkgmanagerd/Get DCOP object
 * It takes care of locking, unlocking, waiting the package
 * manager daemon to be ready, and sets up the various signal
 * connections.
 *
 * You can use it as if it where the real object. If an error
 * occurs, it is reported using the \c error virtual method.
 *
 * @author Sylvain Joyeux
 */

class GetClient : virtual public DCOPObject
{
  K_DCOP

private:
  DCOPRef m_kdedRef, m_getRef;
  bool m_connected;


  /* The kaptmanagerd/Get signals. We
   * use this array to connect to them
   * easily */
  static const QCString s_getSignals[];

  /* Connect the Get signals to this object */
  void connectSignals();
  /* Disconnect us from the Get signals */
  void disconnectSignals();
  /* Reads the Get object error string */
  QString readError();
  /* Calls lock on the KDED pkgmanager daemon.
   * After this call, we wait for the isReady
   * slot to be called */
  bool lock();

  QCString m_pending;
  bool m_pendingHasArg;
  DCOPArg m_pendingArg;
  void setupCall(const QCString& function);
  template<class T>
    void setupCall(const QCString& function, const T& arg);

public:
  GetClient();
  ~GetClient();

  /** Upgrades the package lists */
  bool update();
  /** Do an install/remove on the given packages.
   * @see APT::Get::install for the description
   * of the command syntax */
  bool install(const QStringList& command);
  /** Upgrades the packages
   * @arg distupgrade do a dist-upgrade instead of an upgrade:
   * during a regular upgrade, we aren't allowed to install or
   * remove packages. During dist-upgrades, we can. */
  bool upgrade(bool distupgrade);

  /** When a query has been issued (conffile, stats, ...),
   * the return value should be set by a call to this
   * method. Since the Task objects wait for this answer,
   * if no answer is needed, call it anyway with an empty
   * QVariant as answer value
   *
   * @arg answer the UI return value to the query
   */
  void setQueryAnswer(QVariant answer = QVariant());

  /** Tests if the server is waiting for an answer
   * @return true if we have to call setQueryAnswer, and
   * false otherwise */
  bool needsAnswer();

  /** Cancels the current task
   * @arg wait if true, blocks until the task terminates */
  void cancelCurTask(bool wait);

k_dcop:
  virtual void beginTask() = 0;
  virtual void endTask() = 0;

  virtual void beginStage(int category, QString newstage) = 0;

  virtual void newItem(QString key, QString desc, QString dlinfo, int size) = 0;
  virtual void setItemState(QString key, int state) = 0;
  virtual void itemProgression(QString key, int current, int total) = 0;

  virtual void questionYN(QString message) = 0;
  virtual void questionText(QString message) = 0;

  virtual void conffile(QString file) = 0;

  virtual void stats(APT::Stats stats) = 0;

  virtual void overallProgression(int curItem, int totalItems,
    double curBytes, double totalBytes, double cps) = 0;

  virtual void error(QString key, QString message) = 0;
  virtual void warning(QString key, QString message) = 0;
  virtual void info(QString key, QString message) = 0;

  virtual void subwin(int id) = 0;

  virtual void exit() = 0;

  void pmIsReady();
  void pmBeginTask();
  void pmEndTask();
  void pmExit();
};

#endif
