/***************************************************************************
 *   Copyright (C) 2004 by Sylvain Joyeux                                  *
 *   sylvain.joyeux@m4x.org                                                *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "kdedpkgmanager.h"

#include <kapplication.h>
#include <dcopclient.h>
#include <kprocess.h>
#include <kstandarddirs.h>
#include <qfile.h>
#include <kdesu/client.h>
#include <qfileinfo.h>

#include <kdebug.h>
#include <klocale.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

KDEDPkgManager::KDEDPkgManager(const QCString& obj)
 : KDEDModule(obj), m_daemon(0)
{
  setIdleTimeout(15);
  resetIdle();

  m_daemon = new KProcess;
  connect(m_daemon, SIGNAL(processExited(KProcess*)),
    this, SLOT(pmHasExited(KProcess*)));

  DCOPClient* client = KApplication::dcopClient();
  connect(client, SIGNAL(applicationRegistered(const QCString&)),
      this, SLOT(pmRegistered(const QCString&)));
  connect(client, SIGNAL(applicationRemoved(const QCString&)),
      this, SLOT(lockerRemoved(const QCString&)));
  client -> setNotifications(true);
}

KDEDPkgManager::~KDEDPkgManager()
{
  if (m_daemon->isRunning())
    killDaemon();

  delete m_daemon;
}

bool KDEDPkgManager::lock()
{
  setError(QString::null);

  if (!m_locker.isEmpty())
  {
    kdDebug() << "PkgManager locking by " << callingDcopClient() -> senderId() <<
        " error : already locked by " << m_locker << endl;
    setError(i18n("The package manager daemon is already used by another application"));
    return false;
  }

  if (m_daemon->isRunning())
  {
    m_locker = callingDcopClient() -> senderId();
    emit pmIsReady();
    return true;
  }

  if (launchDaemon())
  {
    m_locker = callingDcopClient() -> senderId();
    return true;
  }

  return false;
}

void KDEDPkgManager::unlockPM()
{
  m_locker = QString::null;
  resetIdle();
}

void KDEDPkgManager::unlock()
{
  QString caller = callingDcopClient() -> senderId();
  if (m_locker != caller)
  {
    kdDebug() << "Locker and unlocker are different ("
      << m_locker << " != " << caller << ")" << endl;
    return;
  }

  unlockPM();
}

bool KDEDPkgManager::isAvailable()
{ return m_locker.isEmpty(); }

bool KDEDPkgManager::hasError()
{ return !m_error.isEmpty(); }
QString KDEDPkgManager::error()
{ return m_error; }
void KDEDPkgManager::setError(const QString err_string)
{ m_error = err_string; }

void KDEDPkgManager::idle()
{
  if (!m_locker.isEmpty())
    resetIdle();
  else if (m_daemon->isRunning())
  {
    kdDebug() << "Stopping kpkgmanagerd" << endl;
    killDaemon();
  }
}

DCOPRef KDEDPkgManager::pmGet()
{ return DCOPRef("kpkgmanagerd", "Get"); }

QCString KDEDPkgManager::getDCOPServer()
{
  QCString dcopSrv;
  QString file = DCOPClient::dcopServerFile();

  // directly from DCOPClient::attachInternal
  QString fName = DCOPClient::dcopServerFile();
  QFile f(fName);
  if (!f.open(IO_ReadOnly)) {
    setError( QString("Internal error: unable to open the DCOP authority file").arg(fName) );
    return 0;
  }
  int size = QMIN( 1024, f.size() ); // protection against a huge file
  QCString contents( size+1 );
  if ( f.readBlock( contents.data(), size ) != size )
  {
    qDebug("Error reading from %s, didn't read the expected %d bytes", fName.latin1(), size);
    // Should we abort ?
  }
  contents[size] = '\0';
  int pos = contents.find('\n');
  if ( pos == -1 ) // Shouldn't happen
  {
      qDebug("Only one line in dcopserver file !: %s", contents.data());
      dcopSrv = QString::fromLatin1(contents);
  }
  else
  {
      dcopSrv = QString::fromLatin1(contents.left( pos ));
  }

  return dcopSrv;
}


// Checks that the current user can execute @p file
// as root without using kdesu (that is, it is setuid root
// and we have the rights to execute it
static bool can_run_asis(const QString& file)
{
  struct stat daemon_stat;
  if (!stat(file.local8Bit(), &daemon_stat)) return false;
  if (!(daemon_stat.st_mode & S_ISUID)) return false;
  if (!(daemon_stat.st_uid == 0)) return false;

  QFileInfo fileinfo(file);
  return fileinfo.isExecutable();
}

bool KDEDPkgManager::launchDaemon()
{
  if (m_daemon -> isRunning())
  {
    emit pmIsReady();
    return true;
  }

  QString daemon = KStandardDirs::findExe("kpkgmanagerd");
  if (daemon.isEmpty())
  {
    setError(i18n("The package manager daemon (kpkgmanagerd) does not seem to be installed on"
                  " your system. Please check your installation"));
    return false;
  }

  QCString local_dcop = getDCOPServer();
  if (local_dcop.isEmpty())
    return false;

  m_daemon -> clearArguments();
  m_daemon -> setEnvironment("DCOPSERVER", local_dcop);

  // Check execution rights on daemon
  // if it is suid root, we launch it as-is
  // otherwise, we use kdesu
  if (!can_run_asis(daemon))
    *m_daemon << "kdesu" << "--nonewdcop" << "--caption" << i18n("Launching KPkgManager daemon") << "--";

  *m_daemon << daemon << "--nofork";

  if (!m_daemon -> start())
  {
    setError(i18n("Unable to launch the package manager daemon (%1)").arg(daemon));
    return false;
  }

  return true;
}


void KDEDPkgManager::killDaemon()
{
  if (!m_daemon->isRunning()) return;

  DCOPClient* client = KApplication::dcopClient();
  client -> send("kpkgmanagerd", "Get", "quit()", QByteArray());
  m_daemon ->wait();
}

extern "C" {
  KDEDModule *create_pkgmanager(const QCString &obj)
  {
     return new KDEDPkgManager(obj);
  }
};

void KDEDPkgManager::pmHasExited( KProcess * )
{
  kdDebug() << "The package manager daemon exited" << endl;
  if (m_locker.isEmpty()) return;

  unlockPM();
  emit pmExit();
}

void KDEDPkgManager::pmRegistered( const QCString & appId )
{
  if (appId == "kpkgmanagerd")
  {
    kdDebug() << "Package manager daemon just registered with DCOP" << endl;
    emit pmIsReady();
  }
}

void KDEDPkgManager::lockerRemoved( const QCString& appId )
{
  if (m_locker != QString(appId)) return;

  kdDebug() << "The locker app has unregistered with DCOP" << endl;

  if (m_daemon->isRunning())
    pmGet().callExt("cancelCurTask", DCOPRef::UseEventLoop, 500, false);

  unlockPM();
}

void KDEDPkgManager::cancel(bool wait)
{
  // brute force cancel
  m_daemon->kill();
}

#include "kdedpkgmanager.moc"
