/***************************************************************************
 *   Copyright (C) 2004 by Sylvain Joyeux                                  *
 *   sylvain.joyeux@m4x.org                                                *
 *
 *   Parts of this code, and some ideas, are from the Apt command line tools
 *   Apt is copyright 1997, 1998, 1999 Jason Gunthorpe and others.
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "acqprogress.h"

#include "task.h"
#include "progress.h"
#include "tasks/common.h"

#include <apt-pkg/acquire-item.h>
#include <apt-pkg/acquire-worker.h>
#include <apt-pkg/error.h>

#include <qstring.h>
#include <qregexp.h>
#include <klocale.h>
#include <kdebug.h>

namespace APT
{

  AcqProgress::AcqProgress(Task* parent)
      : m_parent(parent), m_id(1), m_stop(false) {}


  AcqProgress::~AcqProgress() {}

  void AcqProgress::Reset()
  {
    m_stop = false;
    m_id = 1;
    m_errors.clear();
  }

  void AcqProgress::Start()
  {
    beginStage(m_parent, BeginStage::PreDownload, i18n("Downloading..."));
  }

  bool AcqProgress::MediaChange(string Media, string Drive)
  {
    QString msg = i18n("Media change : please insert the disc labeled\n"
                       " '%1'\n"
                       "in the drive '%2' and press enter")
                  .arg(QString::fromLatin1(Media.c_str()))
                  .arg(QString::fromLatin1(Drive.c_str()));
    return questionyn(m_parent, msg);
  }

  QString AcqProgress::NewItem(pkgAcquire::ItemDesc &Itm)
  {
    if (Itm.Owner->ID)
      return m_keys[Itm.Owner->ID];

    QString description = QString::fromLatin1(Itm.Description.c_str());
    QString dlinfo;
    int id = m_id++;
    Itm.Owner->ID = id;

    QString key;
    static const QRegExp
        rx_lists("(?:Sources|Packages|Release)");

    // Split the description field
    QStringList d = QStringList::split(' ', description);
    if (rx_lists.exactMatch(d.last()))
    { // We are matching a list download. The key is the ID
      description = *d.at(1) + ' ' + *d.at(2) + '\n' + *d.at(0);
      key = QString::number(id);
    }
    else
    { // We are matching a package download. The key is the package name
      // description is made of 4 fields
      key = InstallPackages::getKey(Itm);
      description = *d.at(2) + " " + *d.at(3);
      dlinfo = *d.at(0) + " " + *d.at(1);
    }

    m_keys[id] = key;

    m_parent -> sendInfo(new APT::NewItem(key, description, dlinfo, Itm.Owner->FileSize));
    return key;
  }
  void AcqProgress::IMSHit(pkgAcquire::ItemDesc &Itm)
  {
    QString key = NewItem(Itm);
    m_parent -> sendInfo(new SetItemState(key, SetItemState::Hit));
  }

  void AcqProgress::Fetch(pkgAcquire::ItemDesc &Itm)
  {
    QString key = NewItem(Itm);
    m_parent -> sendInfo(new SetItemState(key, SetItemState::Get));
  }
  void AcqProgress::Done(pkgAcquire::ItemDesc &Itm)
  {
    QString key = NewItem(Itm);
    if (!m_errors.contains(key))
      m_parent -> sendInfo(new SetItemState(key, SetItemState::Downloaded));
  }
  void AcqProgress::Fail(pkgAcquire::ItemDesc &Item)
  {
    if (Item.Owner->Status == pkgAcquire::Item::StatIdle)
      return;

    if (Item.Owner->Status == pkgAcquire::Item::StatDone)
    {
      QString key = NewItem(Item);
      m_parent -> sendInfo(new SetItemState(key, SetItemState::Ignore));
    }
    else
    {
        QString msg(QString::fromLatin1(Item.Owner->ErrorText.c_str()));
      QString key = NewItem(Item);
      m_errors[key] = true;
      m_parent -> sendInfo(new Notify(Notify::Error, msg, key));
    }
  }
  void AcqProgress::Stop()
  {
    Reset();
  }

  void AcqProgress::Cancel()
  {
    QMutexLocker lock(&m_mtx_stop);
    m_stop = true;
  }

  bool AcqProgress::Pulse(pkgAcquire *Owner)
  {
    if (m_stop)
    {
      QMutexLocker lock(&m_mtx_stop);
      m_stop = false;
      return false;
    }

    pkgAcquireStatus::Pulse(Owner);

    m_parent->sendInfo(
        new OverallProgression(CurrentItems, TotalItems, CurrentBytes, TotalBytes, CurrentCPS)
    );

    for (pkgAcquire::Worker *I = Owner->WorkersBegin(); I != 0; I = Owner->WorkerStep(I))
    {
      // There is no item running
      if (I->CurrentItem == 0)
          continue;

      int id = I->CurrentItem->Owner->ID;
      if (!m_keys.contains(id)) continue;
      QString key = m_keys[id];

      m_parent->sendInfo(new ItemProgression(key, I->CurrentSize, I->TotalSize));
    }

    return true;
  }

};

