/***************************************************************************
 *   Copyright (C) 2004 by Sylvain Joyeux                                  *
 *   sylvain.joyeux@m4x.org                                                *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "get.h"

#include "progress.h"

#include "tasks/update.h"
#include "tasks/install.h"
#include "tasks/upgrade.h"

namespace APT {

template<class EventType>
static EventType* convert_to(QCustomEvent* event)
{
  if (event -> type() != EventType::Type) return 0;
  EventType* e = dynamic_cast<EventType *>(event);
  return e;
}

template<class EventType>
static bool match1(QCustomEvent* event, Get* object,
  void (Get::*method)(EventType*))
{
  EventType* e = convert_to<EventType>(event);
  if (!e) return false;

  (object->*method)(e);
  return true;
}

GetPrivate::GetPrivate( Get * parent )
  : m_parent(parent) {}

void GetPrivate::customEvent(QCustomEvent* event)
{
  /* First, the non-query messages */
  if ( match1(event, m_parent, &Get::newItem)
    || match1(event, m_parent, &Get::setItemState)
    || match1(event, m_parent, &Get::itemProgression)) return;

  if ( match1(event, m_parent, &Get::overallProgression) ) return;

  if ( match1(event, m_parent, &Get::beginTask)
      || match1(event, m_parent, &Get::endTask) ) return;

  Notify* notify = convert_to<Notify>(event);
  if (notify)
  {
    if (notify -> level == Notify::Error)
      m_parent->error(notify);
    else if (notify -> level == Notify::Warning)
      m_parent->warning(notify);
    else if (notify -> level == Notify::Info)
      m_parent->info(notify);
    return;
  }


  /* After this line come the query messages */
  m_query = event;
  if (match1(event, m_parent, &Get::conffile)
      || match1(event, m_parent, &Get::subwin)) return;

  if (match1(event, m_parent, &Get::stats)) return;
  if (match1(event, m_parent, &Get::beginStage)) return;

  Prompt* prompt = convert_to<Prompt>(event);
  if (prompt)
  {
    if (prompt -> type == Prompt::QuestionYN)
      m_parent->questionYN(prompt);
    else if (prompt -> type == Prompt::QuestionText)
      m_parent->questionText(prompt);
    return;
  }
}

Get::Get()
{
  m_dispatcher = new GetPrivate(this);
}

Get::~Get()
{
  delete m_dispatcher;
}

bool Get::install( const QStringList & query )
{
  Install* task = new Install(query, m_dispatcher);
  task -> start();
  return true;
}

bool Get::update( )
{
  Update* task = new Update(m_dispatcher);
  task -> start();
  return true;
}

bool Get::upgrade(bool distupgrade)
{
  Upgrade* task = new Upgrade(distupgrade, m_dispatcher);
  task -> start();
  return true;
}

void Get::setQueryAnswer(const QVariant& answer)
{
  QEvent* query = m_dispatcher -> m_query;
  m_dispatcher->m_query = 0;
  Task::setQueryAnswer(query, answer);
}
bool Get::needsAnswer()
{ return m_dispatcher->m_query; }

void Get::beginTask( BeginTask* /*task*/ ) {}
void Get::endTask( EndTask* /*task*/ ) {}
void Get::beginStage( BeginStage* /*event*/ ) {}
void Get::newItem( NewItem * /*event*/ ) {}
void Get::setItemState( SetItemState * /*event*/ ) {}
void Get::itemProgression( ItemProgression * /*event*/ ) {}

void Get::questionYN( Prompt* /*message*/ ) {}
void Get::questionText( Prompt* /*message*/ ) {}

void Get::conffile( Conffile * /*event*/ )
{ setQueryAnswer(QVariant()); }

void Get::subwin( SubwinId* /*event*/ )
{ setQueryAnswer(QVariant()); }

void Get::stats( Stats * /*event*/ ) {}
void Get::overallProgression( OverallProgression * /*event*/ ) {}

void Get::error( Notify* /*event*/ ) {}
void Get::warning( Notify* /*event*/ ) {}
void Get::info( Notify* /*event*/ ) {}

};
