/***************************************************************************
 *   Copyright (C) 2004 by Sylvain Joyeux                                  *
 *   sylvain.joyeux@m4x.org                                                *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#ifndef APTGET_H
#define APTGET_H

#include <qobject.h>

namespace APT
{
  class BeginTask;
  class EndTask;
  class BeginStage;
  class NewItem;
  class SetItemState;
  class ItemProgression;
  class Conffile;
  class Stats;
  class Notify;
  class Prompt;
  class OverallProgression;
  class SubwinId;

  class Get;

  /** @internal
   * This class is needed if we want QObject-derived
   * objects to be able to derive from Get. It catches
   * the messages and forwards dispatches them
   * in the methods of the associated Get object.
   */
  class GetPrivate : public QObject
  {
    Q_OBJECT

    friend class Get;

    Get* m_parent;
    QEvent* m_query;

    GetPrivate(Get* parent);

    /** Event dispatch.
     * This method dispatches known events to the various functions
     * in m_parent.
     * It takes also care to set reply using Task::setQueryAnswer if
     * needed */
    void customEvent(QCustomEvent* event);
  };

  /** Base class giving access to apt-get functionalities.
   * The only thing you need to do is subclass it and implement
   * the virtual methods.
   *
   * @author Sylvain Joyeux
   */
  class Get
  {
    friend class GetPrivate;

  protected:
    /** Since we don't want Get to inherit QObject
     * (because, if we do, we won't be able to derive
     * a class from Get and from a QWidget at the same
     * time), \c m_dispatcher dispatches the event for
     * Get */
    GetPrivate* m_dispatcher;

  public:
    Get();
    virtual ~Get();

    /** Install and removes packages
     * The commands in query are
     *    [+-]packagename(=<version>|/<release>)?
     * where the +- means install or remove
     * the =version selects a particular version
     * the /release selects a particular release
     * no part after the package name selects the
     * version with the highest priority
     *
     * Unlike in apt-get, there is no regexp support
     * for the package name */
    bool install(const QStringList& query);

    /** Does apt-get update */
    bool update();

    /** Does apt-get (dist-)upgrade */
    bool upgrade(bool distupgrade);

  protected:
    void setQueryAnswer(const QVariant& answer);
    bool needsAnswer();

    /** Called just before when the task enters Task::task() */
    virtual void beginTask(BeginTask* task);
    /** Called just after the task left Task::task() */
    virtual void endTask(EndTask* task);

    /** Called when we begin a new stage */
    virtual void beginStage(BeginStage* event);

    /** A new item is being downloaded. Note that
     * this method shouldn't presume of the new item state,
     * since setItemState will be called just after */
    virtual void newItem(NewItem* event);
    /** Sets the download state of an item. The item
     * has already been defined by newItem */
    virtual void setItemState(SetItemState* event);
    /** Gives progression information on an item download */
    virtual void itemProgression(ItemProgression* event);

    /** Needs a Yes/No prompt. If the answer is yes, use setQueryAnswer(true),
     * and false otherwise. */
    virtual void questionYN(Prompt* event);

    /** Needs a prompt whose answer is a phrase (for instance,
     * a dialog asking the user to type <<Yes, do as I say!>>).
     * Set the answer using setQueryAnswer */
    virtual void questionText(Prompt* event);

    /** Asks the user what we want to do with a configuration
     * that has been modified by him and the package maintainer
     * at the same time.
     *
     * You have to call setQueryAnswer() when you're done with
     * the conffile handling
     */
    virtual void conffile(Conffile* event);

    /** Gives informations about the installation which is about to
     * begin. Particularly, it tells us what packages will be installed
     * and removed.
     *
     * Answer false if you want to stop the install process here,
     * and true otherwise. Use setQueryAnswer for setting the answer */
    virtual void stats(Stats* event);

    /** Gives informations about the download progression */
    virtual void overallProgression(OverallProgression* event);

    /** Sends a system-wide or item-wide error
     * @param id the key used to identify the item this message
     *  refers to, or QString::null if this is system-wide.
     *  If the error is system-wide, the task has to finish because of it.
     *  Otherwise, the task can continue its work if it is appropriate */
    virtual void error(Notify* event);
    /** Sends a system-wide or item-wide warning
     * @param id the key used to identify the item this message
     *  refers to, or QString::null if this is system-wide.
     *  This kind of message is non-fatal. */
    virtual void warning(Notify* event);
    /** Gives an information
     * @param id the key used to identify the item this message
     *  refers to, or QString::null if this is system-wide */
    virtual void info(Notify* event);

    /** The package manager needs us to embed a X window (for instance
     * because there is a debconf configuration going on) */
    virtual void subwin(SubwinId* event);
  };
};

#endif
