/***************************************************************************
 *   Copyright (C) 2004 by Sylvain Joyeux                                  *
 *   sylvain.joyeux@m4x.org                                                *
 *
 *   Parts of this code, and some ideas, are from the Apt command line tools
 *   Apt is copyright 1997, 1998, 1999 Jason Gunthorpe and others.
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "packagemanager.h"

#include "task.h"
#include "progress.h"
#include "kprocessrunner.h"

#include <apt-pkg/configuration.h>
#include <apt-pkg/error.h>

#include <klocale.h>

#include <qstringlist.h>
#include <qregexp.h>
#include <qsocketnotifier.h>

#include <kdebug.h>
#include <klocale.h>

namespace APT
{

  PackageManager::PackageManager(pkgDepCache* cache)
      : pkgDPkgPM(cache)
  {
  }

  PackageManager::~PackageManager()
  {}

  /* For APT < 0.6 compatibility */
  bool PackageManager::Go()
  { return Go(0); }
  bool PackageManager::Go( int status_fd )
  {
    Task* parent = Task::currentTask();

    unsigned int MaxArgs = _config->FindI("Dpkg::MaxArgs",350);
    unsigned int MaxArgBytes = _config->FindI("Dpkg::MaxArgBytes",8192);

    /*
    if (RunScripts("DPkg::Pre-Invoke") == false)
      return false;

    if (RunScriptsWithPkgs("DPkg::Pre-Install-Pkgs") == false)
      return false;
    */

    PackageManagerProcess dpkg;
    bool err = false;

    using std::vector;
    for (vector<Item>::iterator I = List.begin(); I != List.end();)
    {
      QStringList args;

      vector<Item>::iterator J = I;
      for (; J != List.end() && J->Op == I->Op; J++);

      // Generate the argument list
      if (J - I > (signed)MaxArgs)
        J = I + MaxArgs;

      // Stick in any custom dpkg options
      Configuration::Item const *Opts = _config->Tree("DPkg::Options");
      if (Opts != 0)
      {
        Opts = Opts->Child;
        for (; Opts != 0; Opts = Opts->Next)
        {
          if (Opts->Value.empty() == true)
            continue;
          args += Opts->Value.c_str();
        }
      }

      switch (I->Op)
      {
      case Item::Remove:
        args += "--force-depends";
        args += "--force-remove-essential";
        args += "--remove";
        break;

      case Item::Purge:
        args += "--force-depends";
        args += "--force-remove-essential";
        args += "--purge";
        break;

      case Item::Configure:
        args += "--configure";
        break;

      case Item::Install:
        args += "--unpack";
        break;
      }

      // Calculate the actual command line size
      uint cmdline_size = args.join(" ").length();

      // Write in the file or package names
      if (I->Op == Item::Install)
      {
        for (;I != J && cmdline_size < MaxArgBytes; I++)
        {
          if (I->File[0] != '/')
          {
            error(parent,
              QString("Internal Error, Pathname to install is not absolute '%1'").arg(I->File.c_str()));
            return false;
          }
          QString file(I->File.c_str());
          args += file;
          cmdline_size += file.length();
        }
      }
      else
      {
        for (;I != J && cmdline_size < MaxArgBytes; I++)
        {
          QString name(I->Pkg.Name());
          args += name;
          cmdline_size += name.length();
        }
      }
      J = I;

      if (0 < status_fd)
      {
          args += "--status-fd";
        args += QString::number(status_fd);
      }

      if (!dpkg.start(args))
      {
          err = true;
        break;
      }
    }


//     if (RunScripts("DPkg::Post-Invoke") == false)
//       return false;
    return !err;

  }






  PackageManagerProcess::PackageManagerProcess()
    : m_parent(0), m_drop(false), m_state(-1)
  {
    QString working_directory(_config->FindDir("DPkg::Run-Directory","/").c_str());
    setWorkingDirectory(working_directory);
    setEnvironment("DPKG_NO_TSTP", "yes");
    setEnvironment("LC_ALL", "C");
    setEnvironment("DEBIAN_FRONTEND", "kde");
    setEnvironment("DEBCONF_KDE_OPTIONS", "embed");
		setUsePty(Stdout, false);

    connect(this, SIGNAL(receivedStdout(KProcess*, char*, int)),
        this, SLOT(receive(KProcess*, char*, int)));
    connect(this, SIGNAL(receivedStderr(KProcess*, char*, int)),
        this, SLOT(receive(KProcess*, char*, int)));
    connect(this, SIGNAL(wroteStdin(KProcess*)),
        this, SLOT(sent(KProcess*)));
  }

  bool PackageManagerProcess::start(const QStringList& args)
  {
    m_parent = Task::currentTask();

    clearArguments();

    m_path = _config->Find("Dir::Bin::dpkg","dpkg").c_str();
    (*this) << m_path << args;

    // Run the subprocess in the GUI thread (needed if we want to get the
    // stdin events properly).
    m_mtxRunner.lock();
    m_runner = new KProcessRunner(this);
    bool is_running = m_runner -> start(KProcess::NotifyOnExit, KProcess::All);
    m_mtxRunner.unlock();

    if (is_running) m_runner -> wait();
    delete m_runner; m_runner = 0;
    if (!is_running) return false;

    // Go back to NullState
    kdDebug() << "End of package manager, going back to NullState" << endl;
    setState(NullState, QString::null);

    // Check for an error code.
    if (normalExit() && exitStatus())
    {
      // @todo RunScripts("DPkg::Post-Invoke");
      return false;
    }
    else if (!normalExit())
      return false;

    return true;
  }




  bool PackageManagerProcess::configureAll()
  {
    QStringList args;
    args += "--configure";
    args += "-a";

    return start(args);
  }





  void PackageManagerProcess::resetState()
  { setState(NullState, QString::null); }
  void PackageManagerProcess::addMessage(const QString& message)
  {
    if (m_state == NullState) return;
    info(m_parent, message, m_package);
  }
  void PackageManagerProcess::setState(int new_state, const QString& pkg)
  {
    // pkg can be either a package name or a file name
    // In the latter case, we need to extract the package name
    static QRegExp rx_package ("([^/_]+)_([^_]+)_.*\\.deb$");
    QString new_package, details;
    if (rx_package.search(pkg) >= 0)
    {
      new_package = rx_package.cap(1);
      details = rx_package.cap(2);
    }
    else
      new_package = pkg;

    if (m_state == new_state && new_package == m_package) return;

    // Let's treat the old package first
    if (!m_package.isEmpty() && m_package != new_package)
    {
      if (m_state == SetItemState::Remove)
        m_parent->sendInfo(new SetItemState(m_package, SetItemState::Removed));
      else if (m_state == SetItemState::Purge)
        m_parent->sendInfo(new SetItemState(m_package, SetItemState::Removed));
      else if (m_state == SetItemState::Setup)
        m_parent->sendInfo(new SetItemState(m_package, SetItemState::Installed));
    }

    // Then, the new one
    if (new_state == NullState)
    {
      // Empty the input buffer
      if (!m_inputBuffer.isEmpty())
      {
        parseInput(m_inputBuffer);
        m_inputBuffer = QString::null;
      }

      kdDebug() << "[Package manager] state change to NullState" << endl;
    }
    else
    {
      kdDebug() << "[Package manager] state change to " << new_state << " for " << new_package << endl;
      if (!m_items.contains(new_package))
      {
        m_items[new_package] = true;
        m_parent->sendInfo(new NewItem(new_package, new_package, details, 0));
      }
      m_parent->sendInfo(new SetItemState(new_package, (SetItemState::ItemState) new_state));
    }

    m_package = new_package;
    m_state = new_state;
  };

  void PackageManagerProcess::parseInput(const QString& line)
  {
    /***************************************************************
     * We define the regular expressions needed to parse dpkg output
     */

    static QRegExp rx_winid("MAINWINDOW_ID (\\d+)");

    static QRegExp rx_conf[] = {
      QRegExp("^Configuration file `([^']+)'"),
      QRegExp("\\(Y/I/N/O/D/Z\\)"),
      QRegExp()
    };

    /* These are the messages that are to be ignored,
     * but tells us that the last task (install, remove,
     * configure, ...) is finished
     * Moreover, when one of these messages is encountered,
     * we drop everything until we get a line matching one
     * of the rx_itemstate regexps */
    static QRegExp rx_ignore[] = {
      QRegExp("Preparing to replace"),
      QRegExp("Reading database"),
      QRegExp("Errors were encountered"),
      QRegExp("Selecting previously deselected"),
      QRegExp("Preconfiguring packages"),
      QRegExp()
    };
    /* These are to be dropped, without any state change */
    static QRegExp rx_drop[] = {
      QRegExp("^Use of uninitialized value in .* at /usr/share/perl5/Debconf"),
      QRegExp()
    };

    struct StateMatch
    {
      int state;
      QRegExp regexp;
    };

    // The order is important, since we use the same as in SetItemState !
    static StateMatch rx_itemstate[] = {
      { SetItemState::Unpack, QRegExp("^Unpacking ([^\\s]+) \\(from (.*)\\)") },
      { SetItemState::Unpack, QRegExp("^Unpacking replacement ([^\\s]+)") },
      { SetItemState::Setup,  QRegExp("^Setting up ([^\\s]+) \\(\\d.*\\) \\.\\.\\.") },
      { SetItemState::Remove, QRegExp("^Removing ([^\\s]+) \\.\\.\\.") },
      { SetItemState::Purge,  QRegExp("^Purging configuration files for ([^\\s]+)") },
      { SetItemState::Error,  QRegExp("error processing ([^\\s]+)") },
      { -1, QRegExp() }
    };
    /***************************************************************/

    kdDebug() << "[Package manager] " << line << endl;

    if (rx_winid.exactMatch(line))
    {
      int id = rx_winid.cap(1).toInt();
      m_parent -> sendInfo(new SubwinId(id));
    }

    for (QRegExp* rx = rx_drop; !rx -> isEmpty(); ++rx)
      if (rx -> search(line) >= 0) return;

    for (QRegExp* rx = rx_ignore; !rx -> isEmpty(); ++rx)
      if (rx -> search(line) >= 0)
      {
        m_drop = true;
        setState(NullState, QString::null);
        return;
      }


    for (StateMatch* match = rx_itemstate; match -> state != -1; ++match)
      if (match -> regexp.search(line) >= 0)
      {
        m_drop = false;

        int new_state = match -> state;
        QString package = match -> regexp.cap(1);
        setState(new_state, package);

        return;
      }


    if (rx_conf[0].search(line) >= 0)
    {
      // Since we are in the GUI thread, we can't block
      // using Task::waitQueryAnswer(). We wait for the
      // hasConffileAnswer slot to be called
      m_drop = true;
      m_parent -> sendQuery(
        new Conffile(rx_conf[0].cap(1)), QVariant(), this, SLOT(hasConffileAnswer(const QVariant&)));
      return;
    }
    else if (rx_conf[1].search(line) >= 0)
    {
      m_drop = false;
    }
    else if (!m_drop)
    {
      kdDebug() << "[Package manager] added messages for " << m_package << endl;
      addMessage(line);
    }
  }


  void PackageManagerProcess::hasConffileAnswer(const QVariant& /*answer*/)
  { send("N\n"); }



  void PackageManagerProcess::receive( KProcess * /*process*/, char * buffer, int len )
  {
    static QRegExp const rx_empty("^\\s*$");
    kdDebug() << "[PackageManager::receive] " << QString::fromLatin1(buffer, len) << endl;
    m_inputBuffer += QString::fromLatin1(buffer, len);
    QStringList lines = QStringList::split('\n', m_inputBuffer, true);

    m_inputBuffer.truncate(0);
    if (!lines.last().isEmpty())
      m_inputBuffer = lines.last();
    lines.pop_back();

    for (QStringList::ConstIterator i = lines.begin(); i != lines.end(); ++i)
    {
        if (rx_empty.exactMatch(*i))
            continue;

        parseInput(*i);
    }
  }

  void PackageManagerProcess::send( const QString & line )
  {
      bool empty = m_pendingOutput.empty();
      m_pendingOutput += line;
    if (empty)
        write();
  }

  void PackageManagerProcess::sent( KProcess * /*proc*/ )
  {
    m_pendingOutput.pop_front();
    write();
  }

  void PackageManagerProcess::write()
  {
      if (m_pendingOutput.empty()) return;

    // let's assure that current_line.latin1() will remain valid
    static QCString current_line;
    current_line = m_pendingOutput.first().local8Bit();

    writeStdin(current_line.data(), current_line.length());
  }

  void PackageManagerProcess::cancel( )
  {
    m_mtxRunner.lock();
    if (m_runner) m_runner->kill();
    m_mtxRunner.unlock();
  }

}

#include "packagemanager.moc"
