/***************************************************************************
 *   Copyright (C) 2004 by Sylvain Joyeux                                  *
 *   sylvain.joyeux@m4x.org                                                *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#ifndef APTPROGRESS_H
#define APTPROGRESS_H

// Here are the various messages that could be sent
// from a Task object.

#include <qevent.h>

#include "task.h"
#include <qstringlist.h>

namespace APT
{
  /** Defines event types used in RTTI */
  enum Types
  {
    BeginTaskEvent = QEvent::User,
    EndTaskEvent,
    BeginStageEvent,

    QuestionYNEvent,
    PromptEvent,

    NewItemEvent,
    SetItemStateEvent,
    ItemProgressionEvent,

    OverallProgressionEvent,
    NotifyEvent,
    StatsEvent,
    ConffileEvent,

    SubwinIdEvent
  };

  /** For debugging purposes only */
  inline QString eventName(QEvent* event)
  {
    static const QString event_names[] =
    {
      "BeginTask", "EndTask", "BeginStage",
      "QuestionYN", "Prompt",
      "NewItem", "SetItemState", "ItemProgression",
      "OverallProgression", "Notify", "Stats", "Conffile",
      "SubwinId"
    };

    return event_names[event -> type() - QEvent::User];
  };

  /** Emitted at the beggining of a new task */
  class BeginTask : public QCustomEvent
  {
  public:
    static const int Type = BeginTaskEvent;

    Task* task;

    BeginTask(Task* task)
      : QCustomEvent(Type), task(task) {}
  };

  /** Emitted when one of the various stages end.
   * @see Stages for the defined stages */
  class EndTask : public QCustomEvent
  {
  public:
    static const int Type = EndTaskEvent;

    Task* task;

    EndTask(Task* task)
      : QCustomEvent(Type), task(task) {}
  };

  /** Emitted when one of the various stages begins.
   * @see Stages for the defined stages */
  class BeginStage : public QCustomEvent
  {
  public:
    static const int Type = BeginStageEvent;

    enum Categories
    {
      PreRun,
      Fix,
      PreDownload,
      PreInstall,
      PostInstall,
      Cancel
    };
    const int category;

    const QString stage;

    BeginStage(int category, QString stage)
      : QCustomEvent(Type), category(category), stage(stage) {}
  };

  inline bool beginStage(Task* task, int category, const QString& stage)
  {
    task->sendQuery(new BeginStage(category, stage), false);
    return task->waitQueryAnswer().toBool();
  };


  /** Asks something from the user.
   * This is a blocking event, so you have to call Task::setQueryAnswer with
   * the actual answer.
   *
   *  A QuestionYN expects a boolean, true if it is yes, and false otherwise.
   *  A QuestionText expects a QString
   */
  class Prompt : public QCustomEvent
  {
  public:
    /** The various prompt styles available */
    enum PromptStyle
    {
      QuestionYN, /// A Y/N question
      QuestionText /// A question that needs a textual answer
    };

    static const int Type = PromptEvent;

    /** The question we ask to the user */
    const QString question;
    /** The prompt style. @see PromptStyle for available values */
    const int type;

    Prompt(const QString& question, int type)
      : QCustomEvent(Type), question(question), type(type) {}
  };
  /** Shortcut for sending a Y/N question from the \arg parent task and
   * getting the answer back
   * @return true if the answer is yes, and false otherwise */
  inline bool questionyn(Task* parent, const QString& question)
  { QEvent* e = new Prompt(question, Prompt::QuestionYN);
    parent -> sendQuery(e, false);
    return parent->waitQueryAnswer().toBool(); }
  /** Shortcut for sending a QuestioText prompt from the \arg parent task and
   * getting the answer back
   * @return the answer or QString::null if the dialog is cancelled */
  inline QString questiontext(Task* parent, const QString& question)
  { QEvent* e = new Prompt(question, Prompt::QuestionText);
    parent -> sendQuery(e, false);
    return parent->waitQueryAnswer().toString(); }




  /** A configuration file has been modified both by the user and
   * the package manager. The current configuration file path is
   * in \c file, and the new one is file + ".dpkg-new". Note that
   * the application HAS TO deal with this problem since the Get
   * PackageManager object will simply say "N" to the dpkg qustion.
   *
   * This is a blocking event, so you HAVE TO call Task::setQueryAnswer
   * with a null QVariant as anwser
   */
  class Conffile : public QCustomEvent
  {
  public:
    static const int Type = ConffileEvent;
    const QString file;

    Conffile(const QString& file)
      : QCustomEvent(Type), file(file) {}
  };




  /** Declares a new item */
  class NewItem : public QCustomEvent
  {
  public:
    static const int Type = NewItemEvent;

    /** The item key. The same key should be used in SetItemState
     * and Notify to refer to this item.
     *
     * QString::null is used in Notify to send global notifications
     * (not package-specific messages)
     *
     * When the items are packages, the package name is used as the key
     * in order to be able to match the items created by the download
     * system with the ones created by the package manager
     */
    const QString key;
    /** A string describing the item */
    const QString description, moreinfo;

    /** When downloading, sets the size of the item */
    const int size;

    NewItem(const QString& key, QString desc, QString more, int size)
      : QCustomEvent(Type), key(key), description(desc), moreinfo(more), size(size) {}
  };

  /** Changes the current state of an item. The item
   * has already been declared with NewItem */
  class SetItemState : public QCustomEvent
  {
  public:
    enum ItemState
    {
      // Downloading states (apt-get)
      Hit,
      Get,
      Downloaded,
      Ignore,
      Error,

      // Package manager states
      // The order is important, as it is used
      // in PackageManager
      Unpack,
      Setup,
      Remove,
      Purge,
      Installed,
      Removed,
      Pending
    };
    static const int Type = SetItemStateEvent;

    const QString key;
    const ItemState state;

    SetItemState(const QString& key, ItemState state)
      : QCustomEvent(Type), key(key), state(state) {}
  };

  /** Sends progression data for an item. The item has
   * been already declared using NewItem */
  class ItemProgression : public QCustomEvent
  {
  public:
    static const int Type = ItemProgressionEvent;

    const QString key;
    const int current, total;

    ItemProgression(const QString& key, int current, int total)
      : QCustomEvent(Type), key(key), current(current), total(total) {}
  };







  /** Sends information about the whole progression */
  class OverallProgression : public QCustomEvent
  {
  public:
    static const int Type = OverallProgressionEvent;

    const int curItem, totalItems;
    const double curBytes, totalBytes;
    const double cps;

    OverallProgression(int currentItem, int totalItems,
      double currentBytes, double totalBytes, double cps)
      : QCustomEvent(Type),
        curItem(currentItem), totalItems(totalItems),
        curBytes(currentBytes), totalBytes(totalBytes),
        cps(cps) {}
  };

  /** Sends a notification. The notification event can
   * be system-wide or specific to an item. On the latter
   * case, \c key contains the key of the item, otherwise it
   * has to be set to QString::null for system-wide messages
   */
  class Notify : public QCustomEvent
  {
  public:
    static const int Type = NotifyEvent;

    enum Levels
    {
      Error,
      Warning,
      Info
    };

    /** The key of the item this notification refers to,
     * or -1 if it is system-wide */
    const QString key;
    /** @see Levels */
    const int level;
    /** The message string */
    const QString message;

    Notify(int level, const QString& msg, const QString& key = QString::null)
      : QCustomEvent(Type), key(key), level(level), message(msg) {}
  };

  /** Sends an error to the \arg parent task in
   * a non-blocking way. */
  inline void error(Task* parent, const QString& msg, const QString& key = QString::null)
  { parent -> sendInfo(new Notify(Notify::Error, msg, key)); };
  /** Sends an warning to the \arg parent task in
   * a non-blocking way. */
  inline void warning(Task* parent, const QString& msg, const QString& key = QString::null)
  { parent -> sendInfo(new Notify(Notify::Warning, msg, key)); };
  /** Sends an information message to the \arg parent task in
   * a non-blocking way. */
  inline void info(Task* parent, const QString& msg, const QString& key = QString::null)
  { parent -> sendInfo(new Notify(Notify::Info, msg, key)); };

  /** Sends APT information about what should be done.
   * Particularly, it conveys the lists of packages */
  class Stats : public QCustomEvent
  {
  public:
    static const int Type = StatsEvent;

    struct DownloadItem
    {
      QString key;
      unsigned long partialSize;
      unsigned long size;

      DownloadItem() : partialSize(0), size(0) {}
      DownloadItem(QString k, unsigned long p, unsigned long s)
        : key(k), partialSize(p), size(s) {}
    };

    typedef QValueList<DownloadItem> Downloads;
    typedef QValueList<DownloadItem>::ConstIterator DownloadIterator;

    QValueList<DownloadItem> downloads;

    void addDownloadItem(const QString& key, unsigned long p, unsigned long s)
    { downloads.append( DownloadItem(key, p, s) ); };

    enum Lists
    {
      Installed = 0,
      Removed,
      Purged,
      Kept,
      Held,
      Upgraded,
      Downgraded,
      Suggested,
      Recommended,
      Essential,
      Infos,
      Errors,
      Warnings,
      Query,
      BadState,

      ListsCount
    };

    QStringList lists[ListsCount];

    double fetch, present, total;
    double unpacked;

    Stats()
      : QCustomEvent(Type),
        fetch(0.0), present(0.0), total(0.0), unpacked(0.0) {}

    const QStringList& get(int idx) const
    { return lists[idx]; }
    QStringList& get(int idx)
    { return lists[idx]; }
    void setByteCounts(double fetch, double present, double total, double unpacked)
    {
      Stats::fetch = fetch;
      Stats::present = present;
      Stats::total = total;
      Stats::unpacked = unpacked;
    }
  };

  inline QDataStream& operator << (QDataStream& stream, const Stats::DownloadItem& item)
  { return stream << item.key << item.partialSize << item.size; }
  inline QDataStream& operator >> (QDataStream& stream, Stats::DownloadItem& item)
  { return stream >> item.key >> item.partialSize >> item.size; }

  inline QDataStream& operator << (QDataStream& stream, const Stats& stats)
  {
    for(int i = 0; i < Stats::ListsCount; ++i)
      stream << stats.lists[i];

    stream << stats.downloads;

    return stream << stats.fetch << stats.present << stats.total << stats.unpacked;
  }
  inline QDataStream& operator >> (QDataStream& stream, Stats& stats)
  {
    for(int i = 0; i < Stats::ListsCount; ++i)
      stream >> stats.lists[i];

    stream >> stats.downloads;

    return stream >> stats.fetch >> stats.present >> stats.total >> stats.unpacked;
  }

  class SubwinId : public QCustomEvent
  {
  public:
    static const int Type = SubwinIdEvent;

    const int id;

    SubwinId(int id)
        : QCustomEvent(Type), id(id) {}
  };


  /*class : public QCustomEvent
  {
  public:
    const int Type = QEvent::User + 1;

    const int id;

    (int id)
      : QCustomEvent(Type), id(id) {}
  };*/

};

#endif
