/***************************************************************************
 *   Copyright (C) 2004 by Sylvain Joyeux                                  *
 *   sylvain.joyeux@m4x.org                                                *
 *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "task.h"

#include "progress.h"
#include "packagesystem.h"
#include "packagemanager.h"

#include <qapplication.h>
#include <qsignal.h>

#include <apt-pkg/init.h>
#include <apt-pkg/error.h>
#include <klocale.h>
#include <kdebug.h>

namespace APT
{
  bool Task::s_init = false;
	QMap<QEvent*, Task*> Task::s_waiting;

  Task::Task(QObject* parent)
      : m_parent(parent),
      m_hasQuery(false), m_queryObject(0), m_querySlot(0),
      m_cancel(false), m_pmanager(0)
  {
    if (!s_init)
    {
      PackageSystem* pkgsys = new PackageSystem;
      bool config = pkgInitConfig(*_config);
      _config -> Set("Apt::System", "KDE-APT");
      bool system = pkgInitSystem(*_config,_system);
      s_init = config && system && (_system == pkgsys);

      if (!s_init)
      {
        sendInfo(new Notify(Notify::Error, i18n("Unable to initialize the packaging system"), QString::null));
        return;
      }
    }
  }


  Task::~Task()
  {
    s_task_mutex.lock();
    if (s_task == this)
      s_task = 0;
    s_task_mutex.unlock();
  }

  bool Task::start()
  {
    if (!s_init) return false;

    {
      QMutexLocker locker(&s_task_mutex);
      // Don't worry, ~Task sets s_task to 0
      // if s_task == this
      if (s_task && s_task -> running())
        return false;

      s_task = this;
    }

    // We clear the pending errors, since most apt-pkg
    // functions fail if _error is not empty
    if (_error->empty())
    	kdDebug() << "[Task::start] Pending errors" << endl;
    while (!_error->empty())
    {
	    string error;
      _error->PopMessage(error);
      QString err_qstring(error.c_str());
      kdDebug() << "[Task::start] " << err_qstring << endl;
    }

    QThread::start();
		return true;
  }

  void Task::run()
  {
    bool go = true;
    sendInfo(new BeginTask(this));

    // Check that dpkg has cleanly finished its
    // last job
    if (PackageSystem::checkUpdates())
    {
      go = questionyn(this,
					i18n("It seems that the package manager has not finished its last job cleanly."
							 " Do you want me to correct this ?"));
      if (go)
      {
				beginStage(this, BeginStage::Fix, "Configuring half-installed packages");
				m_pmanager = new PackageManagerProcess;

        m_pmanager -> configureAll();
				delete m_pmanager;
				m_pmanager = 0;
      }
    }

    if (go) task();

    m_cancel = false;
    if (m_hasQuery)
    {
      m_hasQuery = false;
      QMap<QEvent*, Task *>::Iterator i = s_waiting.begin();
      while (i != s_waiting.end())
      {
        if (i.data() == this) s_waiting.remove(i);
        else ++i;
      }
    }

		sendInfo(new EndTask(this));
  }

  void Task::sendInfo(QEvent* event)
  {
    kdDebug() << "[Task] " << this << " sending info message " << event -> type()
      << " " << eventName(event) << endl;
  	QApplication::postEvent(m_parent, event);
  }

  bool Task::sendQuery(QEvent* event, QVariant cancel, QObject* object, const char* slot)
  {
    if (m_cancel) return false;
    if (m_hasQuery) return false;
    m_hasQuery = true;

    kdDebug() << "[Task] " << this << " sending query message " << event -> type()
      << " " << eventName(event) << endl;

    m_queryCancel = cancel;
    m_queryObject = object;
    m_querySlot = slot;

    s_waiting[event] = this;
    QApplication::postEvent(m_parent, event);
    return true;
  }

  QVariant Task::waitQueryAnswer()
  {
    if (!m_hasQuery) return QVariant();

    m_waitAnswer.wait();
    kdDebug() << "Received query answer !" << endl;
    return m_answer;
  }

  void Task::setQueryAnswer(const QVariant& answer)
  {
		QObject* qobj = 0;
		QString qslot;

		if (!m_hasQuery) return;
		m_answer = answer;

		// allow us to call sendQuery from within
		// the waiting threads
		qobj = m_queryObject;
		qslot = m_querySlot;

		m_hasQuery = false;
		m_waitAnswer.wakeAll();

    if (qobj)
    {

			QSignal* signal = new QSignal;
			signal -> connect(qobj, qslot);
      signal -> setValue(answer);
      signal -> activate();
			delete signal;

      kdDebug() << "setQueryAnswer: Sending signal to " << m_querySlot << endl;
    }
  }

  /*static*/ void Task::setQueryAnswer(QEvent* event, const QVariant& answer)
  {
    if (!s_waiting.contains(event)) return;
  	Task* task = s_waiting[event];
    if (!task) return;
    s_waiting.remove(event);
    task -> setQueryAnswer(answer);
  }

  QMutex Task::s_task_mutex;
  Task* Task::s_task = 0;
  Task* Task::currentTask()
  {
    QMutexLocker locker(&s_task_mutex);
    if (s_task && s_task -> finished())
      s_task = 0;
    return s_task;
  }

  void Task::cancel()
  {
    m_cancel = true;
    kdDebug() << "Generic cancel" << endl;

    if (m_hasQuery)
    {
      kdDebug() << "Canceling current query" << endl;
      setQueryAnswer(m_queryCancel);
    }

		if (m_pmanager)
		{
			kdDebug() << "We are running the package manager only, kill it and return" << endl;
			m_pmanager->cancel();
			return;
		}

    kdDebug() << "Task-specific cancel (cancelTask)" << endl;
    cancelTask();
  }
};

