/***************************************************************************
 *   Copyright (C) 2004 by Sylvain Joyeux                                  *
 *   sylvain.joyeux@m4x.org                                                *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#ifndef APTTASK_H
#define APTTASK_H

#include <qthread.h>
#include <qwaitcondition.h>
#include <qmap.h>
#include <qvariant.h>
#include <qobject.h>

class QEvent;
class QSignal;

namespace APT
{

  class PackageManagerProcess;

  /** Since the apt library provides blocking functions, we
   * have to use multi-threading for keeping the UI alive.
   *
   * This class is the base class the threads should derive from.
   * It provides the simple messaging system used by UI objects. These
   * services are based on the Qt event loop, using QCustomEvent-based
   * messages
   *
   * There are two types of messages :
   *   - non blocking messages, sent via \c sendInfo
   *   - blocking messages, expecting an answer from the GUI. These messages
   *   are sent using sendQuery, the anwser has to be set using Task::setQueryAnswer with
   * the message event as parameter.
   *
   *  The calling object can wait for the answer using waitQueryAnswer, or specify
   * a signal in sendQuery. The first method is to be used when the message
   * is sent from within another thread than the Qt thread. The second one
   * is here to allow using sendQuery from within the Qt thread (for instance,
   * when using KProcess)
   *
   * This system is designed to have only one Task object running at a time
   * (in order to make Task::currentTask possible)
   *
   * @author Sylvain Joyeux
   */

  class Task : public QThread
  {
    /** @internal object to who the task sends its messages */
    QObject* m_parent;

    /** @internal holds the answer after a query has been issued */
    QVariant m_answer, m_queryCancel;
    /** @internal wait flag used to put the thread
     * to sleep and wake it up in sendQuery and setQueryAnswer */
    QWaitCondition m_waitAnswer;
    /** @internal the map of the tasks currently wainting */
    static QMap<QEvent*, Task*> s_waiting;

    /** @internal is the package system initialized or not */
    static bool s_init;
    /** @internal mutex for synchronization on s_task manipulation */
    static QMutex s_task_mutex;
    /** @internal the current task */
    static Task* s_task;

    bool m_hasQuery;
    QObject* m_queryObject;
    const char* m_querySlot;
    void setQueryAnswer(const QVariant& answer);

    /** True if we are currently trying to cancel the task */
    bool m_cancel;

    /** Used when direct access to the package manager is needed */
    PackageManagerProcess* m_pmanager;

  public:
    /** Builds the task object and initializes the package system
     * if needed
     * @arg parent the parent object, that is the object all messages will go to */
    Task(QObject* parent);
    ~Task();

    /** Starts this task. The thread can fail to start
     * mainly for two reasons :
     *      - the package system failed its initialization
     *      - there is already a Task object running (currentTask() is non-null)
     * @return true on success and false on failure
     */
    bool start();

    /** Sends a non-blocking message to the parent object
     * @arg event the event message */
    void sendInfo(QEvent* event);

    /** Sends a query message to the parent object. It returns immediately.
     * Call waitQueryAnswer to wait for the message answer, or wait for
     * us calling the specified slot. Only one query event can be on fly
     * at the same time.
     *
     * The slot to be called is given as a parameter of sendQuery. It
     * is disconnected when the answer arrives. Its signature has to
     * be void slot(const QVariant&)
     *
     * Note that the signal is emitted from within the <em>answering</em>
     * thread whereas waitQueryAnswer stays in the calling thread.
     *
     * @arg event the event message
     * @arg object slot parent, or 0 if you don't want to use the slot system
     * @arg slot the slot to be called
     * @return false if there is already a query event waiting for its answer,
     *  in which case the specified event is not sent, and has to be destroyed
     *  by the caller */

    bool sendQuery(QEvent* event, QVariant cancel, QObject* object = 0, const char * slot = 0);

    /** Waits that a call to setQueryAnswer sets the answer to \c event, and
     * returns this answer. Note that "event" is only used as a key, it does not
     * have to be a pointer to a valid object. */
    QVariant waitQueryAnswer();

    /** Sets the return value to a query message
     * @arg event the query QEvent object
     * @arg answer the return value
     */
  	static void setQueryAnswer(QEvent* event, const QVariant& answer);

    /** Returns the current task. Note that it works if there is
     * only one Task object running at the same time. Task::start
     * checks that it is the case */
    static Task* currentTask();

    /** Terminates this task. Don't overload this method, but cancelTask()
     * instead
     */
    void cancel();

  protected:
    /** Do what is needed to cancel cleanly this task's job */
    virtual void cancelTask() = 0;
    /** The task main method. Overload this instead of run */
    virtual void task() = 0;

    /** This is the standard entry point for QThread. It is overloaded
     * here in order to provide some common initialization. Do not overload
     * it, but task instead */
    virtual void run();
  };

};

#endif
