/***************************************************************************
 *   Copyright (C) 2004 by Sylvain Joyeux                                  *
 *   sylvain.joyeux@m4x.org                                                *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "cachefile.h"

#include "../progress.h"
#include "../task.h"
#include "../packagesystem.h"

#include <apt-pkg/error.h>
#include <apt-pkg/algorithms.h>

#include <klocale.h>

namespace APT
{

CacheFile::CacheFile( Task * parent )
   : m_parent(parent), List(0) {}

bool CacheFile::CheckDeps(bool AllowBroken)
{
  if (_error->PendingError() == true)
    return false;

  // Check that the system is OK
  if (DCache->DelCount() != 0 || DCache->InstCount() != 0)
  {
    error(m_parent, "Internal Error, non-zero counts");
    return false;
  }

  // Apply corrections for half-installed packages
  if (pkgApplyStatus(*DCache) == false)
    return false;

  // Nothing is broken
  if (DCache->BrokenCount() == 0 || AllowBroken == true)
    return true;

  bool fix_broken = _config->FindB("APT::Get::Fix-Broken",false);

  if (fix_broken == false)
  {
    // @todo show broken dependecies to the user
    m_parent->sendQuery(new Prompt(
        i18n("There are unmet dependencies. Do you want me to try harder ?"), Prompt::QuestionYN), false);
    bool result = m_parent->waitQueryAnswer().toBool();
    if (!result)
    {
      error(m_parent, i18n("Unmet dependencies. Try \"apt-get -f install\" in a console"));
      return false;
    }
  }

  // Attempt to fix broken things
  if (pkgFixBroken(*DCache) == false || DCache->BrokenCount() != 0)
  {
    // @todo show broken dependecies to the user
    error(m_parent, i18n("Unable to correct dependencies. Try \"apt-get -f install\" in a console"));
    return false;
  }
  if (pkgMinimizeUpgrade(*DCache) == false)
  {
    error(m_parent, i18n("Unable to minimize the upgrade set. Try \"apt-get -f install\" in a console"));
    return false;
  }

  return true;
}

// CacheFile::NameComp - QSort compare by name        /*{{{*/
// ---------------------------------------------------------------------
/* */
pkgCache *CacheFile::SortCache = 0;
int CacheFile::NameComp(const void *a,const void *b)
{
   if (*(pkgCache::Package **)a == 0 || *(pkgCache::Package **)b == 0)
      return *(pkgCache::Package **)a - *(pkgCache::Package **)b;

   const pkgCache::Package &A = **(pkgCache::Package **)a;
   const pkgCache::Package &B = **(pkgCache::Package **)b;

   return strcmp(SortCache->StrP + A.Name,SortCache->StrP + B.Name);
}
                  /*}}}*/
// CacheFile::Sort - Sort by name          /*{{{*/
// ---------------------------------------------------------------------
/* */
void CacheFile::Sort()
{
   delete [] List;
   List = new pkgCache::Package *[Cache->Head().PackageCount];
   memset(List,0,sizeof(*List)*Cache->Head().PackageCount);
   pkgCache::PkgIterator I = Cache->PkgBegin();
   for (;I.end() != true; I++)
      List[I->ID] = I;

   SortCache = *this;
   qsort(List,Cache->Head().PackageCount,sizeof(*List),NameComp);
}

bool CacheFile::BuildCaches( bool WithLock )
{
  if (PackageSystem::checkUpdates())
  {
    // The inconsistent states in the package system
    // should have been handled by Task::run. Here, we issue
    // an error.
    error(m_parent, i18n("The package manager is in an inconsisten state. This is an internal error."), QString::null);
    return false;
  }

  OpTextProgress Prog(*_config);
  if (pkgCacheFile::BuildCaches(Prog,WithLock) == false)
    return false;
  return true;
}

bool CacheFile::Open( bool WithLock )
{
  if (PackageSystem::checkUpdates())
  {
    // The inconsistent states in the package system
    // should have been handled by Task::run. Here, we issue
    // an error.
    error(m_parent, i18n("The package manager is in an inconsisten state. This is an internal error."), QString::null);
    return false;
  }


  OpTextProgress Prog(*_config);
  if (pkgCacheFile::Open(Prog,WithLock) == false)
    return false;
  Sort();

  return true;
}

bool CacheFile::OpenForInstall( )
{
  if (_config->FindB("Get::Print-URIs") == true)
    return Open(false);
  else
    return Open(true);
}

};

