/***************************************************************************
 *   Copyright (C) 2004 by Sylvain Joyeux                                  *
 *   sylvain.joyeux@m4x.org                                                *
 *                                                                         *
 *   Parts of this code, and some ideas, are from the Apt command line tools
 *   Apt is copyright 1997, 1998, 1999 Jason Gunthorpe and others.
 *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "common.h"

#include "cachefile.h"

#include "../task.h"
#include "../progress.h"
#include "../acqprogress.h"

#include <apt-pkg/error.h>
#include <apt-pkg/pkgrecords.h>
#include <apt-pkg/algorithms.h>
#include <apt-pkg/versionmatch.h>
#include <apt-pkg/depcache.h>
#include <apt-pkg/sourcelist.h>
#include <apt-pkg/acquire-item.h>
#include <apt-pkg/acquire-worker.h>
#include <apt-pkg/sptr.h>
#include <sys/statvfs.h>

#include <kdebug.h>
#include <klocale.h>

namespace APT
{
  InstallPackages::InstallPackages(CacheFile& cache, Task* m_parent)
      : m_parent(m_parent), m_cache(cache), m_stop(false)
  {
    m_acq = new APT::AcqProgress (m_parent);
  }
  InstallPackages::~InstallPackages()
  {
    delete m_acq;
  }


  /** Indicator functions used for building package lists */
  bool select_new(APT::CacheFile& Cache, pkgCache::PkgIterator& I)
  { return Cache[I].NewInstall(); };
  bool select_del(CacheFile& Cache, pkgCache::PkgIterator& I)
  {
    return Cache[I].Delete() == true
      && (Cache[I].iFlags & pkgDepCache::Purge) != pkgDepCache::Purge;
  }
  bool select_purge(CacheFile& Cache, pkgCache::PkgIterator& I)
  { return Cache[I].Delete() == true
      && (Cache[I].iFlags & pkgDepCache::Purge) == pkgDepCache::Purge; }
  bool select_kept(CacheFile& Cache, pkgCache::PkgIterator& I)
  {
    return
         Cache[I].Upgrade() == false
        && Cache[I].Upgradable() == true
        && I->CurrentVer != 0
        && Cache[I].Delete() == false;
  }
  bool select_upgraded(CacheFile& Cache, pkgCache::PkgIterator& I)
  { return Cache[I].Upgrade() == true && Cache[I].NewInstall() == false; }
  bool select_downgraded(CacheFile& Cache, pkgCache::PkgIterator& I)
  { return Cache[I].Downgrade() == true && Cache[I].NewInstall() == false; }

  bool select_held(CacheFile& Cache, pkgCache::PkgIterator& I)
  {
    return
      (Cache[I].InstallVer != (pkgCache::Version *)I.CurrentVer()
      && I->SelectedState == pkgCache::State::Hold);
  }

  bool select_badstates(CacheFile& /*Cache*/, pkgCache::PkgIterator& I)
  {
    return
        (I.State() != pkgCache::PkgIterator::NeedsNothing);
  }

  typedef bool (*SelectFunction)(APT::CacheFile&, pkgCache::PkgIterator& I);

  /** Builds package lists
   * @return the list of packages where F(package) == true */
  template<SelectFunction F>
  QStringList select(APT::CacheFile& Cache)
  {
    QStringList ret;
    for (unsigned J = 0; J < Cache->Head().PackageCount; J++)
    {
      pkgCache::PkgIterator I(Cache,Cache.List[J]);
      if (F(Cache, I))
        ret += I.Name();
    }
    return ret;
  }

  static QStringList select_essentials(CacheFile& Cache)
  {
    QStringList ret;
    for (unsigned J = 0; J < Cache->Head().PackageCount; J++)
    {
      pkgCache::PkgIterator I(Cache,Cache.List[J]);
      bool is_essential =
        ((I->Flags & pkgCache::Flag::Essential) == pkgCache::Flag::Essential)
        || ((I->Flags & pkgCache::Flag::Important) == pkgCache::Flag::Important);
      if (!is_essential)
        continue;

      if (Cache[I].Delete())
      {
        ret += I.Name();
        continue;
      }

      if (I.CurrentVer() == 0) continue;

      QStringList guilty_depends;
      // Print out any essential package depends that are to be removed
      for (pkgCache::DepIterator D = I.CurrentVer().DependsList(); D.end() == false; D++)
      {
        // Skip everything but depends
        if (D->Type != pkgCache::Dep::PreDepends &&
            D->Type != pkgCache::Dep::Depends)
            continue;

        pkgCache::PkgIterator P = D.SmartTargetPkg();
        if (Cache[P].Delete() == true)
          guilty_depends += QString("%1").arg(P.Name());
      }
      if (guilty_depends.empty()) continue;

      ret += i18n("%1 (due to %2)").arg(I.Name()).arg(guilty_depends.join(" "));
    }
    return ret;
  }

  /** Marks to "purge" the packages which are to be removed */
  void InstallPackages::setPurgeDefault()
  {
    pkgCache::PkgIterator I = m_cache->PkgBegin();
    for (; I.end() == false; I++)
    {
      if (I.Purge() == false && m_cache[I].Mode == pkgDepCache::ModeDelete)
        m_cache->MarkDelete(I,true);
    }
  }

  /** Build the various package lists */
  void InstallPackages::buildStats(Stats* stats)
  {
    // @todo Show all the various warning indicators
    stats -> get(Stats::Purged) = select<&select_purge>(m_cache);
    stats -> get(Stats::Removed) = select<&select_del>(m_cache);
    stats -> get(Stats::Kept) = select<&select_kept>(m_cache);
    stats -> get(Stats::Held) = select<&select_held>(m_cache);
    stats -> get(Stats::Installed) = select<&select_new>(m_cache);
    stats -> get(Stats::Upgraded) = select<&select_upgraded>(m_cache);
    stats -> get(Stats::Downgraded) = select<&select_downgraded>(m_cache);
    stats -> get(Stats::BadState) = select<&select_badstates>(m_cache);

    stats -> get(Stats::Essential) = select_essentials(m_cache);
  }


  void InstallPackages::cancel()
  {
    m_acq -> Cancel();

    QMutexLocker lock(&m_mtxCancel);
    m_stop = true;
  }

  bool InstallPackages::must_stop()
  {
    QMutexLocker lock(&m_mtxCancel);
    bool ret = m_stop;
    m_stop = false;
    return ret;
  }

  // InstallPackages - Actually download and install the packages    /*{{{*/
  // ---------------------------------------------------------------------
  /* This displays the informative messages describing what is going to
     happen and then calls the download routines */
  bool InstallPackages::run(Stats* stats)
  {
    beginStage(m_parent, BeginStage::PreDownload, i18n("Preparing package manager work"));

    {
      QMutexLocker lock(&m_mtxCancel);
      m_stop = false;
    }

    if (_config->FindB("APT::Get::Purge",false) == true)
      setPurgeDefault();

    // Sanity check
    if (m_cache->BrokenCount() != 0)
    {
      kdDebug() << "Internal Error, InstallPackages was called with broken packages!" << endl;
      return false;
    }

    // There's nothing to be done. Still, we send an empty stats message
    if (m_cache->DelCount() == 0 && m_cache->InstCount() == 0 &&
        m_cache->BadCount() == 0)
    {
      m_parent->sendQuery(stats, false);
      m_parent->waitQueryAnswer();
      return true;
    }

    // No remove allowed
    if (m_cache->DelCount() != 0 && _config->FindB("APT::Get::Remove",true) == false)
    {
      error(m_parent, i18n("Packages need to be removed but Remove is disabled."));
      return false;
    }

    // Lock the archive directory
    FileFd Lock;
    if (_config->FindB("Debug::NoLocking",false) == false)
    {
      Lock.Fd(GetLock(_config->FindDir("Dir::Cache::Archives") + "lock"));
      if (_error->PendingError() == true)
      {
        error(m_parent, i18n("Unable to lock the download directory"));
        return false;
      }

    }

    // Create the text record parser
    pkgRecords Recs(m_cache);
    if (_error->PendingError() == true)
      return false;

    // Create the download object
    pkgAcquire Fetcher(m_acq);

    // Read the source list
    pkgSourceList List;
    if (List.ReadMainList() == false)
    {
      error(m_parent, i18n("The list of sources could not be read."));
      return false;
    }

    // Create the package manager and prepare to download
    SPtr<pkgPackageManager> PM = _system->CreatePM(m_cache);
    if (PM->GetArchives(&Fetcher, &List, &Recs) == false ||
        _error->PendingError() == true)
      return false;


    // Build install statistics
    buildStats(stats);

    // Get the list of to-be-downloaded packages
    for (pkgAcquire::UriIterator i = Fetcher.UriBegin(); i != Fetcher.UriEnd(); i++)
    {
      pkgAcquire::Item* owner = i->Owner;
      if (owner->Local) continue;

      QString description = QString::fromLatin1(i->Description.c_str());
      QString key = getKey(description);
      stats->addDownloadItem(key, owner->PartialSize, owner->FileSize);
    }

    double FetchBytes = Fetcher.FetchNeeded();
    double FetchPBytes = Fetcher.PartialPresent();
    double DebBytes = Fetcher.TotalNeeded();
    stats -> setByteCounts(FetchBytes, FetchPBytes, DebBytes, m_cache->UsrSize());

    // Sanity check (once again)
    if (_error->PendingError() == true)
      return false;

    /* Check for enough free space, but only if we are actually going to
       download */
    if (_config->FindB("APT::Get::Download",true) == true)
    {
      struct statvfs Buf;
      string OutputDir = _config->FindDir("Dir::Cache::Archives");
      if (statvfs(OutputDir.c_str(),&Buf) != 0)
      {
        error(m_parent, i18n("Couldn't determine free space in %1").arg(OutputDir.c_str()));
        return false;
      }
      if (unsigned(Buf.f_bfree) < (FetchBytes - FetchPBytes)/Buf.f_bsize)
      {
        error(m_parent, i18n("You don't have enough free space in %1.").arg(OutputDir.c_str()));
        return false;
      }
    }

    // Send query and wait for the UI answer
    // The checks on Trivial-Only or Essential packages
    // are to be made by the UI
    if (must_stop())
      return true;

    m_parent->sendQuery(stats, false);
    if (!m_parent -> waitQueryAnswer().toBool())
      return false;

    /* Unlock the dpkg lock if we are not going to be doing an install
       after. */
    if (_config->FindB("APT::Get::Download-Only",false) == true)
      _system->UnLock();

    while (1)
    {
      if (must_stop())
        return true;

      bool Transient = false;
      if (_config->FindB("APT::Get::Download",true) == false)
      {
        for (pkgAcquire::ItemIterator I = Fetcher.ItemsBegin(); I < Fetcher.ItemsEnd();)
        {
          if ((*I)->Local == true)
          {
            I++;
            continue;
          }

          // Close the item and check if it was found in cache
          (*I)->Finished();
          if ((*I)->Complete == false)
            Transient = true;

          // Clear it out of the fetch list
          delete *I;
          I = Fetcher.ItemsBegin();
        }
      }

      if (Fetcher.Run() == pkgAcquire::Failed)
        return false;

      // Print out errors
      bool Failed = false;
      for (pkgAcquire::ItemIterator I = Fetcher.ItemsBegin(); I != Fetcher.ItemsEnd(); I++)
      {
        if ((*I)->Status == pkgAcquire::Item::StatDone &&
            (*I)->Complete == true)
          continue;

        if ((*I)->Status == pkgAcquire::Item::StatIdle)
        {
          Transient = true;
          // Failed = true;
          continue;
        }

        /*fprintf(stderr,_("Failed to fetch %s  %s\n"),(*I)->DescURI().c_str(),
                (*I)->ErrorText.c_str());*/
        Failed = true;
      }

      /* If we are in no download mode and missing files and there were
         'failures' then the user must specify -m. Furthermore, there
         is no such thing as a transient error in no-download mode! */
      if (Transient == true &&
          _config->FindB("APT::Get::Download",true) == false)
      {
        Transient = false;
        Failed = true;
      }

      if (Failed && _config->FindB("APT::Get::Fix-Missing",false) == false)
        return false;

      /** For testing the download part */
      if (_config->FindB("APT::Get::Download-Only",false) == true)
        return true;

      if (Failed && Transient == true)
        return false;

      if (Failed && _config->FindB("APT::Get::Fix-Missing",false) == false)
        return false;

      if (Failed && PM->FixMissing() == false)
        return false;

      _system->UnLock();

      // Run it
      if (!beginStage(m_parent, BeginStage::PreInstall, i18n("Running the package manager ...")))
        return true;

      pkgPackageManager::OrderResult Res = PM->DoInstall();
      if (Res == pkgPackageManager::Failed || _error->PendingError() == true)
        return false;
      if (Res == pkgPackageManager::Completed)
      {
        beginStage(m_parent, BeginStage::PostInstall, i18n("Running post-installation scripts ..."));
        return true;
      }

      // Reload the fetcher object and loop again for media swapping
      Fetcher.Shutdown();
      if (PM->GetArchives(&Fetcher,&List,&Recs) == false)
        return false;

      _system->Lock();
    }

    return true;
  }

//   bool InstallPackages(
//   {
//     bool Fail = false;
//     bool Essential = false;
//
//     if (Essential == true && Saftey == true)
//     {
//       if (_config->FindB("APT::Get::Trivial-Only",false) == true)
//       {
//         error(m_parent, i18n("Trivial Only specified but this is not a trivial operation."));
//         return false;
//       }
//
//       QString prompt = i18n("Yes, do as I say!");
//       QString message = i18n("You are about to do something potentially harmful"
//                              "To continue type in the phrase '%1'");
//
//       /* Annoying prompt for dangerous actions */
//       QString ret =
//         questiontext(m_parent, message.arg(prompt));
//       if (ret.lower() != prompt.lower())
//       {
//         error(m_parent, i18n("Aborted"));
//         return false;
//       }
//     }
//     else
//     {
//       if (_config->FindB("APT::Get::Trivial-Only",false) == true)
//       {
//         error(m_parent, i18n("Trivial Only specified but this is not a trivial operation."));
//         return false;
//       }
//
//       if (_config->FindB("APT::Get::Assume-Yes",false) == false)
//       {
//       }
//     }
//

};
