/***************************************************************************
 *   Copyright (C) 2004 by Sylvain Joyeux                                  *
 *   sylvain.joyeux@m4x.org                                                *
 *
 *   Parts of this code, and some ideas, are from the Apt command line tools
 *   Apt is copyright 1997, 1998, 1999 Jason Gunthorpe and others.
 *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "install.h"

#include "cachefile.h"
#include "common.h"

#include "../progress.h"
#include "../acqprogress.h"

#include <apt-pkg/error.h>
#include <apt-pkg/pkgrecords.h>
#include <apt-pkg/algorithms.h>
#include <apt-pkg/versionmatch.h>
#include <apt-pkg/depcache.h>
#include <apt-pkg/sourcelist.h>
#include <apt-pkg/acquire-item.h>
#include <apt-pkg/acquire-worker.h>
#include <apt-pkg/sptr.h>
#include <sys/statvfs.h>

#include <kdebug.h>
#include <klocale.h>

#include <qregexp.h>

namespace APT
{

static QString TryToChangeVer(pkgCache::PkgIterator Pkg,pkgDepCache &Cache,
                              const char *VerTag,bool IsRel)
{
  pkgVersionMatch Match(VerTag,(IsRel == true?pkgVersionMatch::Release :
                                pkgVersionMatch::Version));

  pkgCache::VerIterator Ver = Match.Find(Pkg);

  if (Ver.end())
    return i18n("Release '%1' for '%2' was not found").arg(VerTag).arg(Pkg.Name());

  Cache.SetCandidateVersion(Ver);
  return QString::null;
}


static bool TryToInstall(pkgCache::PkgIterator Pkg, pkgDepCache &Cache,
                  pkgProblemResolver &Fix, Stats* stats,
                  bool Remove, bool BrokenFix,
                  unsigned int &ExpectedInst)
{
  /* This is a pure virtual package and there is a single available
     provides */
  if (Cache[Pkg].CandidateVer == 0 && Pkg->ProvidesList != 0 &&
      Pkg.ProvidesList()->NextProvides == 0)
  {
    pkgCache::PkgIterator Tmp = Pkg.ProvidesList().OwnerPkg();
    stats -> get(Stats::Installed) += i18n("%1 (instead of %2)").arg(Tmp.Name()).arg(Pkg.Name());
    Pkg = Tmp;
  }

  // Handle the no-upgrade case
  if (_config->FindB("APT::Get::upgrade",true) == false &&
      Pkg->CurrentVer != 0)
  {
    stats -> get(Stats::Kept) += i18n("%1 (upgrade is disabled)").arg(Pkg.Name());
    return true;
  }

  // Check if there is something at all to install
  pkgDepCache::StateCache &State = Cache[Pkg];
  if (Remove == true && Pkg->CurrentVer == 0)
  {
    Fix.Clear(Pkg);
    Fix.Protect(Pkg);
    Fix.Remove(Pkg);

    stats -> get(Stats::Warnings) += i18n("Package %1 is not installed, so not remove").arg(Pkg.Name());
    return true;
  }

  if (State.CandidateVer == 0 && Remove == false)
  {
    stats -> get(Stats::Errors) += i18n("Package %1 has no installation candidate").arg(Pkg.Name());
    return false;
  }

  Fix.Clear(Pkg);
  Fix.Protect(Pkg);
  if (Remove == true)
  {
    Fix.Remove(Pkg);
    Cache.MarkDelete(Pkg,_config->FindB("APT::Get::Purge",false));
    return true;
  }

  // Install it
  Cache.MarkInstall(Pkg,false);
  if (State.Install() == false)
  {
    if (_config->FindB("APT::Get::ReInstall",false) == true)
    {
      if (Pkg->CurrentVer == 0 || Pkg.CurrentVer().Downloadable() == false)
      {
        stats -> get(Stats::Errors)
          += i18n("Reinstallation of %1 is not possible, it cannot be downloaded").arg(Pkg.Name());
      }
      else
        Cache.SetReInstall(Pkg,true);
    }
    else
      stats -> get(Stats::Infos) += i18n("%1 is already the newest version").arg(Pkg.Name());
  }
  else
    ExpectedInst++;

  // Install it with autoinstalling enabled.
  if (State.InstBroken() == true && BrokenFix == false)
    Cache.MarkInstall(Pkg, true);
  return true;
}


/** Use the cache state and the packages the user asked for and fills in the extra package list
 (packages needed to install but not asked for), the suggests and the recommends */
static void sort_packages(CacheFile& Cache, Stats* stats)
{
  /** We know what packages have been queried already (in stats->query).
   * InstallPackages will fill stats->installed and stats->removed. It's
   * the GUI job to show what packages are "extra install" */

  /* Print out a list of suggested and recommended packages */
  for (unsigned J = 0; J < Cache->Head().PackageCount; J++)
  {
    pkgCache::PkgIterator I(Cache,Cache.List[J]);

    /* Just look at the ones we want to install */
    if ((*Cache)[I].Install() == false)
      continue;

    for (pkgCache::VerIterator V = I.VersionList(); V.end() == false; V++)
    {
      for (pkgCache::DepIterator D = V.DependsList(); D.end() == false; D++)
      {
        pkgCache::DepIterator Start;
        pkgCache::DepIterator End;
        D.GlobOr(Start,End);

        /*
        * If this is a virtual package, we need to check the list of
        * packages that provide it and see if any of those are
        * installed
        */
        pkgCache::PrvIterator Prv = Start.TargetPkg().ProvidesList();
        bool providedBySomething = false;
        for (; Prv.end() != true; Prv++)
          if ((*Cache)[Prv.OwnerPkg()].InstVerIter(*Cache).end() == false)
          {
            providedBySomething = true;
            break;
          }

        if (providedBySomething) continue;

        do
        {
          /* Skip if package is  installed already, or is about to be */
          QString target = Start.TargetPkg().Name();
          if ((*Start.TargetPkg()).SelectedState == pkgCache::State::Install
              || Cache[Start.TargetPkg()].Install())
            break;

          /* Skip if we already saw it */
          if (stats->get(Stats::Suggested).contains(target)
            || stats->get(Stats::Recommended).contains(target))
            break;

          if (Start->Type == pkgCache::Dep::Suggests)
            stats -> get(Stats::Suggested) += target;

          if (Start->Type == pkgCache::Dep::Recommends)
            stats -> get(Stats::Recommended) += target;

          if (Start == End)
            break;
          Start++;
        }
        while (1);
      }
    }
  }
}









Install::Install(const QStringList& packages, QObject* parent):
    Task(parent), m_packages(packages),
    m_cancel(false), m_install(0)
{}


Install::~Install()
{}

void Install::task()
{
  beginStage(this, BeginStage::PreDownload, i18n("Computing needed actions ..."));

  CacheFile Cache(this);
  if (Cache.OpenForInstall() == false ||
      Cache.CheckDeps(true) == false)
  {
    error(this, i18n("Could not open the package cache for install"));
    return;
  }

  // Enter the special broken fixing mode if the user specified arguments
  bool BrokenFix = false;
  if (Cache->BrokenCount() != 0)
    BrokenFix = true;

  unsigned int ExpectedInst = 0;
  unsigned int Packages = 0;
  pkgProblemResolver Fix(Cache);

  Stats stats;

  for (QStringList::Iterator i = m_packages.begin(); i != m_packages.end(); ++i)
  {
    static QRegExp rx_command("([-+])([^/=]+)(?:([/=])(.+))?");

    QString package;
    bool Remove = false;
    QString VerTag;
    bool VerIsRel = false;

    if (!rx_command.exactMatch(*i))
    {
      kdDebug() << "Invalid install command : " << *i << endl;
      continue;
    }

    package = rx_command.cap(2);
    QString mode = rx_command.cap(1);
    if (mode == "+")
      Remove = false;
    else if (mode == "-")
      Remove = true;

    mode = rx_command.cap(3);
    if (mode == "/")
    {
      VerIsRel = true;
      VerTag = rx_command.cap(4);
    }
    else if (mode == "=")
    {
      VerIsRel = false;
      VerTag = rx_command.cap(4);
    }

    // Locate the package
    pkgCache::PkgIterator Pkg = Cache->FindPkg(package.latin1());
    Packages++;
    if (Pkg.end() == true)
    {
      error(this, i18n("Couldn't find package %1").arg(package));
      return;
    }

    if (!VerTag.isEmpty())
    {
      QString ret = TryToChangeVer(Pkg,Cache,VerTag,VerIsRel);
      if (!ret.isNull())
      {
        error(this, ret);
        return;
      }
    }

    if (TryToInstall(Pkg, Cache, Fix, &stats, Remove, BrokenFix, ExpectedInst) == false)
      return;
    else
      kdDebug() << "Updated cache for " << (Remove ? "removing " : "installing ") << package << endl;


    stats.get(Stats::Query) += package;

    if (m_cancel) return;
  }

  /* If we are in the Broken fixing mode we do not attempt to fix the
     problems. This is if the user invoked install without -f and gave
     packages */
  if (BrokenFix == true && Cache->BrokenCount() != 0)
  {
    error(this, i18n("Unmet dependencies. Try 'apt-get -f install' with no packages in a console"));
    // @todo show broken dependencies to the user
    // ShowBroken(c1out,Cache,false);

    return;
  }

  // Call the scored problem resolver
  Fix.InstallProtect();
  if (Fix.Resolve(true) == false)
    _error->Discard();

  if (m_cancel) return;

  // Now we check the state of the packages,
  if (Cache->BrokenCount() != 0)
  {
    QString error_msg =
      i18n("Some packages could not be installed. This may mean that you have "
           "requested an impossible situation or if you are using the unstable "
           "distribution that some required packages have not yet been created "
           "or been moved out of Incoming.");

    if (Packages == 1)
    {
      error_msg = error_msg + "\n" +
                  i18n("Since you only requested a single operation it is extremely likely that "
                       "the package is simply not installable and a bug report against "
                       "that package should be filed.");
    }

    error(this, error_msg);

    // @todo show broken dependencies to the user
    return;
  }

  sort_packages(Cache, &stats);

  Stats* stats_msg = new Stats(stats);
  {
    QMutexLocker locker(&m_mtxCancel);
    m_install = new InstallPackages(Cache, this);
  }
  m_install -> run(stats_msg);
}


void Install::cancelTask()
{
  beginStage(this, BeginStage::Cancel, i18n("Stopping at user request ..."));

  QMutexLocker lock(&m_mtxCancel);;
  m_cancel = true;
  if (m_install) m_install -> cancel();
}

};

