/***************************************************************************
 *   Copyright (C) 2004 by Sylvain Joyeux                                  *
 *   sylvain.joyeux@m4x.org                                                *
 *
 *   Parts of this code, and some ideas, are from the Apt command line tools
 *   Apt is copyright 1997, 1998, 1999 Jason Gunthorpe and others.
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "update.h"

#include "cachefile.h"

#include "../progress.h"

#include <apt-pkg/error.h>
#include <apt-pkg/init.h>
#include <apt-pkg/sourcelist.h>
#include <apt-pkg/cachefile.h>
#include <apt-pkg/sourcelist.h>
#include <apt-pkg/acquire-item.h>
#include <apt-pkg/acquire-worker.h>

#include <klocale.h>
#include <kdebug.h>

namespace APT
{

Update::Update(QObject* parent)
  : Task(parent), m_acq(this)
{
}


Update::~Update()
{
}

void Update::task()
{
  // Get the source list
  pkgSourceList List;
  if (List.ReadMainList() == false)
  {
    error(this, i18n("Error reading sources list"));
    return;
  }

  // Lock the list directory
  FileFd Lock;
  if (_config->FindB("Debug::NoLocking",false) == false)
  {
    Lock.Fd(GetLock(_config->FindDir("Dir::State::Lists") + "lock"));
    if (_error->PendingError() == true)
    {
      error(this, i18n("Unable to lock the list directory"));
      return;
    }
  }

  // Create the download object
  m_acq.Reset();
  pkgAcquire Fetcher(&m_acq);

  // Populate it with the source selection
  if (List.GetIndexes(&Fetcher) == false)
    return;

  // Run it
  if (Fetcher.Run() == pkgAcquire::Failed)
    return;

  for (pkgAcquire::ItemIterator I = Fetcher.ItemsBegin(); I != Fetcher.ItemsEnd(); I++)
  {
    if ((*I)->Status == pkgAcquire::Item::StatDone)
      continue;

    (*I)->Finished();
  }

  // Clean out any old list files
  if (_config->FindB("APT::Get::List-Cleanup",true) == true)
  {
    if (Fetcher.Clean(_config->FindDir("Dir::State::lists")) == false ||
        Fetcher.Clean(_config->FindDir("Dir::State::lists") + "partial/") == false)
    {
      error(this, i18n("Error clearing old list files"));
      return;
    }
  }

  // Prepare the cache.
  CacheFile Cache(this);
  if (Cache.BuildCaches() == false)
  {
    error(this, i18n("Error building caches"));
    return;
  }
}

void Update::cancelTask()
{
  kdDebug() << "Cancel for Update" << endl;
  beginStage(this, BeginStage::Cancel, i18n("Stopping at user request ..."));
  m_acq.Cancel();
}

};
